__all__ = ["CAST_FUNCTION_NAMES", "generic_calls_fix_component"]

from ioshelper.base.reloadable_plugin import OptimizersComponent

from .generic_calls_fix import CAST_FUNCTIONS, generic_calls_fix_optimizer_t

generic_calls_fix_component = OptimizersComponent.factory("Generic calls fixer", [generic_calls_fix_optimizer_t])

CAST_FUNCTION_NAMES = list(CAST_FUNCTIONS.values())



__all__ = ["CAST_FUNCTIONS", "generic_calls_fix_optimizer_t"]

import re

import ida_hexrays
from ida_hexrays import mblock_t, mcallarg_t, mcallargs_t, mcallinfo_t, minsn_t, minsn_visitor_t
from ida_typeinf import tinfo_t
from idahelper import cpp, tif
from idahelper.microcode import mcallarg, minsn, mop

from ioshelper.base.utils import CounterMixin, match_dict

CAST_FUNCTIONS: dict[str | re.Pattern, str] = {
    "OSMetaClassBase::safeMetaCast": "OSDynamicCast",
    "__ZN15OSMetaClassBase12safeMetaCastEPKS_PK11OSMetaClass": "OSDynamicCast",
    "OSMetaClassBase::requiredMetaCast": "OSRequiredCast",
    "__ZN15OSMetaClassBase16requiredMetaCastEPKS_PK11OSMetaClass": "OSDynamicCast",
}

ALLOC_FUNCTION: dict[str | re.Pattern, str] = {
    "OSObject_typed_operator_new": "OSObjectTypeAlloc",
    "_OSObject_typed_operator_new": "OSObjectTypeAlloc",
}

OS_OBJECT_TYPE: tinfo_t = tif.from_c_type("OSObject*")  # type: ignore  # noqa: PGH003
SIZE_T_TYPE: tinfo_t = tif.from_c_type("size_t")  # type: ignore  # noqa: PGH003


class insn_optimizer_t(minsn_visitor_t, CounterMixin):
    def visit_minsn(self) -> int:
        # We only want calls
        insn: minsn_t = self.curins
        if insn.opcode == ida_hexrays.m_call:
            self.visit_call_insn(insn, self.blk)
        return 0

    def visit_call_insn(self, insn: minsn_t, blk: mblock_t):
        res = self.try_convert_cast(insn)
        if not res:
            self.try_convert_alloc(insn)

    def try_convert_alloc(self, insn: minsn_t) -> bool:
        name = minsn.get_func_name_of_call(insn)
        if name is None or (new_name := match_dict(ALLOC_FUNCTION, name)) is None:
            return False

        # Verify call info
        call_info = self.get_call_info(insn, 2)
        if call_info is None:
            return True

        # Get the kty
        kty_name = self.get_arg_name(call_info, 0)
        if kty_name is None or not kty_name.endswith("_kty"):
            # No name, cannot optimize
            return True
        kty_name = kty_name[:-4]

        # Get the class
        cls_type = tif.from_struct_name(kty_name)
        if cls_type is None:
            print(f"[Error] Failed to get type for class: {kty_name}")
            return True

        self.modify_call(cls_type, new_name, insn, call_info, 0, SIZE_T_TYPE)
        return True

    def try_convert_cast(self, insn: minsn_t) -> bool:
        name = minsn.get_func_name_of_call(insn)
        if name is None or (new_name := match_dict(CAST_FUNCTIONS, name)) is None:
            return False

        # Verify call info
        call_info = self.get_call_info(insn, 2)
        if call_info is None:
            return True

        # Convert name to type
        cls_name_mangled = self.get_arg_name(call_info, 1)
        if cls_name_mangled is None:
            # No name, cannot optimize
            return True

        cls_name = cpp.demangle_class_only(cls_name_mangled)
        if cls_name is None:
            print(f"[Error] Failed to extract class name: {cls_name_mangled}")
            return True

        # Get the class
        cls_type = tif.from_struct_name(cls_name)
        if cls_type is None:
            print(f"[Error] Failed to get type for class: {cls_name}")
            return True

        self.modify_call(cls_type, new_name, insn, call_info, 1, OS_OBJECT_TYPE)
        return True

    def get_call_info(self, call_insn: minsn_t, required_args_count: int) -> mcallinfo_t | None:
        """Get call info of the given call instruction, verifying it has the required args count"""
        call_info: mcallinfo_t | None = call_insn.d.f
        if call_info is None or len(call_info.args) != required_args_count:
            return None
        return call_info

    def get_arg_name(self, call_info: mcallinfo_t, arg_index: int) -> str | None:
        """Get the name of the {arg_index} argument, assuming it is a global one"""
        arg: mcallarg_t = call_info.args[arg_index]
        if arg.t != ida_hexrays.mop_a:
            # not const
            return None

        return mop.get_name(arg.a)

    def modify_call(
        self,
        cls_type: tinfo_t,
        new_name: str,
        insn: minsn_t,
        call_info: mcallinfo_t,
        arg_to_remove: int,
        single_arg_type: tinfo_t,
    ) -> None:
        # Assumes size is 2

        cls_type_pointer = tif.pointer_of(cls_type)

        # Check if already handled
        if call_info.return_type == cls_type_pointer:
            return

        # Apply name and type changes
        insn.l.make_helper(f"{new_name}<{cls_type.get_type_name()}>")
        call_info.return_type = cls_type_pointer

        # Remove metaclass arg
        args: mcallargs_t = call_info.args
        if arg_to_remove == 0:
            args[0].swap(args[1])

        args.pop_back()
        call_info.solid_args -= 1

        # Remove the name associated with the first parameter, so there will be no inlay hint
        new_arg = mcallarg.from_mop(call_info.args[0], single_arg_type)
        call_info.args.pop_back()
        call_info.args.push_back(new_arg)

        self.count()


class generic_calls_fix_optimizer_t(ida_hexrays.optinsn_t):
    def func(self, blk: mblock_t, ins: minsn_t, optflags: int):
        # Let IDA reconstruct the calls before
        if blk.mba.maturity < ida_hexrays.MMAT_CALLS:
            return 0

        insn_optimizer = insn_optimizer_t(blk.mba, blk)
        ins.for_all_insns(insn_optimizer)
        return insn_optimizer.cnt



__all__ = [
    "BlockArgByRefField",
    "BlockByRefArgBaseFieldsAssignments",
    "create_block_arg_byref_type",
    "get_block_byref_args_lvars",
    "is_block_arg_byref_type",
]

from dataclasses import dataclass
from enum import Enum

import ida_hexrays
from ida_hexrays import cexpr_t, cfunc_t, cinsn_t, lvar_t
from ida_typeinf import tinfo_t
from idahelper import tif
from idahelper.ast import cexpr

from .utils import StructFieldAssignment

PTR_SIZE = 8
INT_SIZE = 4

OFFSET_ISA = 0
OFFSET_FORWARDING = OFFSET_ISA + PTR_SIZE
OFFSET_FLAGS = OFFSET_FORWARDING + PTR_SIZE
OFFSET_SIZE = OFFSET_FLAGS + INT_SIZE
OFFSET_BYREF_KEEP = OFFSET_SIZE + INT_SIZE
OFFSET_BYREF_DISPOSE = OFFSET_BYREF_KEEP + PTR_SIZE
OFFSETS = [OFFSET_ISA, OFFSET_FORWARDING, OFFSET_FLAGS, OFFSET_SIZE, OFFSET_BYREF_KEEP, OFFSET_BYREF_DISPOSE]


class BlockArgByRefField(Enum):
    """
    struct _block_byref_x {
        void *isa;
        struct _block_byref_x *forwarding;
        int flags;
        int size;
        /* optional */ void (*byref_keep)(void  *dst, void *src);
        /* optional */ void (*byref_dispose)(void *);
        typeof(marked_variable) marked_variable;
    };
    """

    ISA = 0
    FORWARDING = 1
    FLAGS = 2
    SIZE = 3
    HELPER_KEEP = 4
    HELPER_DISPOSE = 5
    VARIABLE = 6

    def get_offset(self, has_helpers: bool | None = None) -> int:
        if self == BlockArgByRefField.VARIABLE and has_helpers is None:
            raise ValueError("has_helpers must be specified for VARIABLE")

        if self == BlockArgByRefField.VARIABLE:
            return OFFSET_BYREF_DISPOSE + PTR_SIZE if has_helpers else OFFSET_BYREF_KEEP

        return OFFSETS[self.value]


TYPE_BLOCK_ARG_BYREF_PREFIX = "Block_byref_layout_"
TYPE_DECL_BLOCK_ARG_BYREF_WITH_HELPERS = """
#pragma pack(push, 1)
struct {class_name} {{
    void *isa;
    struct {class_name} *forwarding;
    int flags;
    int size;
    void (*byref_keep)(void  *dst, void *src);
    void (*byref_dispose)(void *);
    {var_type} value;
}};
#pragma pack(pop)
"""
TYPE_DECL_BLOCK_ARG_BYREF_WITHOUT_HELPERS = """
#pragma pack(push, 1)
struct {class_name} {{
    void *isa;
    struct {class_name} *forwarding;
    int flags;
    int size;
    {var_type} value;
}};
#pragma pack(pop)
"""


def get_type_name_for_addr(ea: int) -> str:
    return f"{TYPE_BLOCK_ARG_BYREF_PREFIX}{ea:08X}"


def is_block_arg_byref_type(tinfo: tinfo_t) -> bool:
    """Check if the given `tif` is a block arg byref type"""
    if not tinfo.is_struct():
        return False
    # noinspection PyTypeChecker
    name: str | None = tinfo.get_type_name()
    return name is not None and name.startswith(TYPE_BLOCK_ARG_BYREF_PREFIX)


def create_block_arg_byref_type(ea: int, var_size: int, has_helpers: bool) -> tinfo_t:
    """Create a tinfo_t for the block byref type"""
    class_name = get_type_name_for_addr(ea)

    if (existing_type := tif.from_struct_name(class_name)) is not None:
        return existing_type

    var_type = tif.from_size(var_size).dstr()
    if has_helpers:
        type_decl = TYPE_DECL_BLOCK_ARG_BYREF_WITH_HELPERS.format(class_name=class_name, var_type=var_type)
    else:
        type_decl = TYPE_DECL_BLOCK_ARG_BYREF_WITHOUT_HELPERS.format(class_name=class_name, var_type=var_type)

    tif.create_from_c_decl(type_decl)
    return tif.from_struct_name(class_name)


def get_block_byref_args_lvars(func: cfunc_t) -> list[lvar_t]:
    """Get all Obj-C block by ref args variables in the function"""
    return [lvar for lvar in func.get_lvars() if is_block_arg_byref_type(lvar.type())]


@dataclass
class BlockByRefArgBaseFieldsAssignments:
    assignments: list[cinsn_t]
    ea: int | None
    type: tinfo_t | None
    isa: cexpr_t | None
    forwarding: cexpr_t | None
    flags: cexpr_t | None
    size: cexpr_t | None
    byref_keep: cexpr_t | None
    byref_dispose: cexpr_t | None

    def __str__(self) -> str:
        return (
            f"isa: {self.isa.dstr() if self.isa else 'None'}, "
            f"forwarding: {self.forwarding.dstr() if self.forwarding else 'None'}, "
            f"flags: {self.flags.dstr() if self.flags else 'None'}, "
            f"size: {self.size.dstr() if self.size else 'None'}, "
            f"byref_keep: {self.byref_keep.dstr() if self.byref_keep else 'None'}, "
            f"byref_dispose: {self.byref_dispose.dstr() if self.byref_dispose else 'None'}, "
        )

    @staticmethod
    def initial() -> "BlockByRefArgBaseFieldsAssignments":
        return BlockByRefArgBaseFieldsAssignments(
            assignments=[],
            ea=None,
            type=None,
            isa=None,
            forwarding=None,
            flags=None,
            size=None,
            byref_keep=None,
            byref_dispose=None,
        )

    def is_completed(self) -> bool:
        """Check if all base fields have been assigned"""
        return self.isa is not None and self.forwarding is not None and self.flags is not None and self.size is not None

    def add_assignment(self, assignment: StructFieldAssignment) -> bool:
        """Add an assignment to the list of assignments"""
        if self.type is None:
            self.type = assignment.type

        field_name = assignment.member.name
        if field_name == "isa":
            self.isa = assignment.expr
            self.ea = assignment.insn.ea
        elif field_name == "forwarding":
            self.forwarding = assignment.expr
        elif field_name == "flags":
            if assignment.is_cast_assign:
                # We need to split it to flags and reserved
                expr = assignment.expr
                if expr.op != ida_hexrays.cot_num:
                    print(f"[Error] invalid flags assignment. Expected const, got: {expr.dstr()}")
                    return False

                num_val = expr.numval()
                self.flags = cexpr.from_const_value(num_val & 0xFF_FF_FF_FF, is_hex=True)
                self.size = cexpr.from_const_value(num_val >> 32, is_hex=True)
            else:
                self.flags = assignment.expr
        elif field_name == "size":
            self.size = assignment.expr
        elif field_name == "byref_keep":
            self.byref_keep = assignment.expr
        elif field_name == "byref_dispose":
            self.byref_dispose = assignment.expr
        else:
            return False

        self.assignments.append(assignment.insn)
        return True



import ida_idaapi
import idaapi
from ida_idaapi import plugin_t

from ioshelper.base.reloadable_plugin import PluginCore, ReloadablePlugin


# noinspection PyPep8Naming
class iOSHelperPlugin(ReloadablePlugin):
    flags = ida_idaapi.PLUGIN_HIDE
    wanted_name = "iOS helper"
    wanted_hotkey = ""
    comment = "Optimize iOS patterns in the code"
    help = ""

    def __init__(self):
        # Use lambda to plugin_core, so it could be fully reloaded from disk every time.
        # noinspection PyTypeChecker
        super().__init__("ioshelper", "ioshelper", plugin_core_wrapper_factory, extra_packages_to_reload=["idahelper"])


def plugin_core_wrapper_factory(*args, **kwargs) -> PluginCore:
    # Reload the module
    idaapi.require("ioshelper.core")
    # Bring the module into locals
    import ioshelper.core

    return ioshelper.core.plugin_core(*args, **kwargs)


# noinspection PyPep8Naming
def PLUGIN_ENTRY() -> plugin_t:
    return iOSHelperPlugin()



__all__ = ["jump_to_string_component"]

import ida_kernwin
import idaapi
from ida_kernwin import action_handler_t

from ioshelper.base.reloadable_plugin import UIAction, UIActionsComponent

from .jump_to_string import jump_to_string_ask

ACTION_ID = "ioshelper:jump_to_string"

jump_to_string_component = UIActionsComponent.factory(
    "Jump to function using a specific string",
    [
        lambda core: UIAction(
            ACTION_ID,
            idaapi.action_desc_t(ACTION_ID, "Jump to function using a specific string", JumpToStringAction(), "Ctrl+S"),
            menu_location=UIAction.base_location(core),
        )
    ],
)


class JumpToStringAction(action_handler_t):
    def activate(self, ctx: ida_kernwin.action_ctx_base_t):
        jump_to_string_ask()
        return 0

    def update(self, ctx):
        return idaapi.AST_ENABLE_ALWAYS



__all__ = [
    "FLAG_BLOCK_HAS_COPY_DISPOSE",
    "BlockBaseFieldsAssignments",
    "block_member_is_arg_field",
    "get_block_type",
    "get_ida_block_lvars",
    "is_block_type",
]

from dataclasses import dataclass

import ida_hexrays
from ida_hexrays import (
    cexpr_t,
    cfuncptr_t,
    cinsn_t,
    lvar_t,
)
from ida_typeinf import tinfo_t, udm_t
from idahelper.ast import cexpr

from .utils import StructFieldAssignment

IDA_BLOCK_TYPE_NAME_PREFIX = "Block_layout_"
IDA_BLOCK_TYPE_BASE_FIELD_NAMES = {
    "isa",
    "flags",
    "reserved",
    "invoke",
    "descriptor",
}

FLAG_BLOCK_HAS_COPY_DISPOSE = 1 << 25


def get_ida_block_lvars(func: cfuncptr_t) -> list[lvar_t]:
    """Get all Obj-C block variables in the function"""
    return [lvar for lvar in func.get_lvars() if is_block_type(lvar.type())]


def is_block_type(tinfo: tinfo_t) -> bool:
    """Check if the type is an Obj-C block type"""
    if not tinfo.is_struct():
        return False
    # noinspection PyTypeChecker
    name: str | None = tinfo.get_type_name()
    return name is not None and name.startswith(IDA_BLOCK_TYPE_NAME_PREFIX)


def block_member_is_arg_field(udm: udm_t) -> bool:
    """Check if the member is an argument field of an Obj-C block"""
    return udm.name not in IDA_BLOCK_TYPE_BASE_FIELD_NAMES


BLOCK_TYPES: dict[str, str] = {}
for typ in ["stack", "global", "malloc", "auto", "finalizing", "weak"]:
    typ_cap = typ.capitalize()
    BLOCK_TYPES[f"_NSConcrete{typ_cap}Block"] = typ
    BLOCK_TYPES[f"__NSConcrete{typ_cap}Block"] = typ
    BLOCK_TYPES[f"__NSConcrete{typ_cap}Block_ptr"] = typ
    BLOCK_TYPES[f"_OBJC_CLASS_$___NS{typ_cap}Block__"] = typ


def get_block_type(isa: str) -> str:
    """Get the block type from the isa symbol"""
    return BLOCK_TYPES.get(isa, "unknown")


@dataclass
class BlockBaseFieldsAssignments:
    assignments: list[cinsn_t]
    ea: int | None
    type: tinfo_t | None
    isa: cexpr_t | None
    flags: cexpr_t | None
    reserved: cexpr_t | None
    invoke: cexpr_t | None
    descriptor: cexpr_t | None

    def __str__(self) -> str:
        return (
            f"isa: {self.isa.dstr() if self.isa else 'None'}, "
            f"flags: {self.flags.dstr() if self.flags else 'None'}, "
            f"reserved: {self.reserved.dstr() if self.reserved else 'None'}, "
            f"invoke: {self.invoke.dstr() if self.invoke else 'None'}, "
            f"descriptor: {self.descriptor.dstr() if self.descriptor else 'None'}"
        )

    @staticmethod
    def initial() -> "BlockBaseFieldsAssignments":
        return BlockBaseFieldsAssignments(
            assignments=[], isa=None, flags=None, reserved=None, invoke=None, descriptor=None, type=None, ea=None
        )

    def is_completed(self) -> bool:
        """Check if all base fields have been assigned"""
        return (
            self.isa is not None
            and self.flags is not None
            and self.reserved is not None
            and self.invoke is not None
            and self.descriptor is not None
        )

    def add_assignment(self, assignment: StructFieldAssignment) -> bool:
        """Add an assignment to the list of assignments"""
        if self.type is None:
            self.type = assignment.type

        field_name = assignment.member.name
        if field_name == "isa":
            self.isa = assignment.expr
            self.ea = assignment.insn.ea
        elif field_name == "flags":
            if assignment.is_cast_assign:
                # We need to split it to flags and reserved
                expr = assignment.expr
                if expr.op != ida_hexrays.cot_num:
                    print(f"[Error] invalid flags assignment. Expected const, got: {expr.dstr()}")
                    return False

                num_val = expr.numval()
                self.flags = cexpr.from_const_value(num_val & 0xFF_FF_FF_FF, is_hex=True)
                self.reserved = cexpr.from_const_value(num_val >> 32, is_hex=True)
            else:
                self.flags = assignment.expr
        elif field_name == "reserved":
            self.reserved = assignment.expr
        elif field_name == "invoke":
            self.invoke = assignment.expr
        elif field_name == "descriptor":
            self.descriptor = assignment.expr
        else:
            return False

        self.assignments.append(assignment.insn)
        return True



from collections.abc import Callable
from dataclasses import dataclass

import ida_hexrays
from ida_hexrays import lvar_t, mba_t, mcallarg_t, minsn_t, mop_t
from ida_typeinf import tinfo_t
from idahelper import memory, tif
from idahelper.microcode import mop
from idahelper.microcode.visitors import extended_microcode_visitor_t

from ..generic_calls_fix import CAST_FUNCTION_NAMES


@dataclass(frozen=True)
class HelperXref:
    name: str


@dataclass(frozen=True)
class FuncXref:
    ea: int


@dataclass(frozen=True)
class IndirectCallXref:
    type: tinfo_t
    offset: int


SourceXref = HelperXref | FuncXref | IndirectCallXref
CallCallback = Callable[["Call", object], None]


@dataclass(frozen=True)
class XrefsMatcher:
    helpers: dict[str, CallCallback]
    """Mapping between helper names and their callbacks"""
    calls: dict[int, CallCallback]
    """Mapping between function addresses and their callbacks"""
    indirect_calls: dict[int, list[tuple[int, CallCallback]]]
    """Mapping between indirect call offset -> type tid -> callback"""
    untyped_indirect_call_fallback: CallCallback | None
    """Callback for indirect call on untyped object"""

    def match_indirect_call(self, typ: tinfo_t, offset: int) -> CallCallback | None:
        if (not typ.is_ptr() and not typ.is_struct()) or (typ.is_ptr() and not typ.get_pointed_object().is_struct()):
            return self.untyped_indirect_call_fallback

        candidates = self.indirect_calls.get(offset)
        if candidates is None:
            return None
        if typ.is_ptr():
            typ = typ.get_pointed_object()

        type_parent_tifs = tif.get_parent_classes(typ)
        if type_parent_tifs is None:
            return None
        type_parent_tifs.append(typ)

        type_parents = {parent.get_tid() for parent in type_parent_tifs}
        for candidate_type, callback in candidates:
            if candidate_type in type_parents:
                return callback

        return None

    @staticmethod
    def build(
        callbacks: list[tuple[SourceXref, CallCallback]], untyped_indirect_callback: CallCallback | None = None
    ) -> "XrefsMatcher":
        """Build a matcher from the given callbacks."""
        helpers = {}
        calls = {}
        indirect_calls = {}
        for xref, callback in callbacks:
            if isinstance(xref, HelperXref):
                helpers[xref.name] = callback
            elif isinstance(xref, FuncXref):
                calls[xref.ea] = callback
            elif isinstance(xref, IndirectCallXref):
                indirect_calls.setdefault(xref.offset, []).append((xref.type.get_tid(), callback))

        return XrefsMatcher(helpers, calls, indirect_calls, untyped_indirect_callback)


def process_function_calls(func_mba: mba_t, matcher: XrefsMatcher, ref: object):
    """Get all calls to the given function in the current function."""
    StaticCallExtractorVisitor(matcher, ref).visit_function(func_mba)


# TODO: This is a hack, we should find a way to remove all consts
CHAR_POINTER_TYPES: list[tinfo_t] = [
    tif.from_c_type("char*"),
    tif.from_c_type("const char*"),
    tif.from_c_type("char* const"),
]  # type: ignore  # noqa: PGH003
VOID_POINTER_TYPE: tinfo_t = tif.from_c_type("void*")  # type: ignore  # noqa: PGH003
ParsedParam = str | int | None


@dataclass
class IndirectCallInfo:
    target_mop: mop_t
    """The target operand of the indirect call"""
    offset: int
    """The offset in the vtable of the call"""
    var: lvar_t | None = None
    """If called on a var, the var"""
    field: tuple[tinfo_t, int] | None = None
    """If called on a field, the type and the offset"""


@dataclass
class Call:
    """A call to a function"""

    func_ea: int
    """The ea of the function"""
    ea: int
    """The ea of the call instruction"""
    params_op: list[mcallarg_t]
    """The operands of the call instruction"""
    params: list[ParsedParam]
    """Parsed params of the call instruction"""
    params_names: list[str | None]
    """If the parameter is a global variable (or reference to it), return its name"""
    assignee: mop_t | None
    """The assignee of this call instruction"""
    indirect_info: IndirectCallInfo | None
    """Info about the indirect call"""

    def __str__(self):
        params_str = ", ".join([
            repr(param) if param is not None else op_param.dstr()
            for param, op_param in zip(self.params, self.params_op)  # noqa: B905
        ])
        assignee_str = f", assignee={self.assignee.dstr()}" if self.assignee is not None else ""
        return f"Call(ea={hex(self.ea)}, params=[{params_str}]{assignee_str}"


class StaticCallExtractorVisitor(extended_microcode_visitor_t):
    def __init__(self, matcher: XrefsMatcher, ref: object):
        super().__init__()
        self.matcher = matcher
        self.has_indirect_calls = bool(matcher.indirect_calls)
        self.has_direct_calls = bool(matcher.calls) or bool(matcher.helpers)
        self.ref = ref

    def _visit_insn(self, ins: minsn_t) -> int:
        if ins.opcode == ida_hexrays.m_call and self.has_direct_calls:
            self._visit_call(ins)
            return 0
        elif ins.opcode == ida_hexrays.m_icall:  # and self.has_indirect_calls:
            self._visit_icall(ins)
            return 0

        return 0

    def _visit_call(self, ins: minsn_t):
        """Search for direct call and invoke the callback."""
        if ins.l.t == ida_hexrays.mop_v:
            callback = self.matcher.calls.get(ins.l.g)
        elif ins.l.t == ida_hexrays.mop_h:
            callback = self.matcher.helpers.get(ins.l.helper)
        else:
            return

        if callback is None:
            return

        callback(self._build_call_for_callback(ins, indirect_info=None), self.ref)

    def _visit_icall(self, ins: minsn_t):  # noqa: C901
        # Search for indirect call of x->vtable->func or x->field->vtable->func
        # Which is x => *x => *x->vtable (offset 0) => *x->vtable + offsetOf(func) => *x->vtable->func
        # Expects ldx
        if ins.r.t != ida_hexrays.mop_d or ins.r.d.opcode != ida_hexrays.m_ldx:
            return

        # Expects add
        ldx_ins = ins.r.d
        if ldx_ins.r.t != ida_hexrays.mop_d or ldx_ins.r.d.opcode != ida_hexrays.m_add:
            return
        add_ins = ldx_ins.r.d

        # Expects add with const
        const_offset = mop.get_const_int(add_ins.r)
        if const_offset is None:
            return

        # Except ldx
        if add_ins.l.t != ida_hexrays.mop_d or add_ins.l.d.opcode != ida_hexrays.m_ldx:
            return
        ldx_ins_2 = add_ins.l.d

        # Maybe expect local
        if ldx_ins_2.r.t == ida_hexrays.mop_l:
            lvar = ldx_ins_2.r.l.var()
            lvar_type = lvar.type()
            callback = self.matcher.match_indirect_call(lvar_type, const_offset)
            indirect_info = IndirectCallInfo(ldx_ins_2.r, const_offset, var=lvar)
        # Maybe expect ldx of an add with const (field)
        elif ldx_ins_2.r.t == ida_hexrays.mop_d and ldx_ins_2.r.d.opcode == ida_hexrays.m_ldx:
            ldx_ins_3 = ldx_ins_2.r.d

            # Expect add with const
            if ldx_ins_3.r.t != ida_hexrays.mop_d or ldx_ins_3.r.d.opcode != ida_hexrays.m_add:
                return

            add_ins_2 = ldx_ins_3.r.d
            const_offset_2 = mop.get_const_int(add_ins_2.r)
            if const_offset_2 is None:
                return

            # Expect local
            if add_ins_2.l.t != ida_hexrays.mop_l:
                return

            lvar = add_ins_2.l.l.var()
            lvar_type: tinfo_t = lvar.type()
            if lvar_type.is_ptr():
                lvar_type = lvar_type.get_pointed_object()
            field_type = self._try_get_member_type(lvar_type, const_offset_2)
            callback = self.matcher.match_indirect_call(field_type, const_offset)
            indirect_info = IndirectCallInfo(ldx_ins_2.r, const_offset, field=(lvar_type, const_offset_2))
        else:
            return

        if callback is None:
            return

        callback(self._build_call_for_callback(ins, indirect_info), self.ref)

    @staticmethod
    def _try_get_member_type(typ: tinfo_t, offset: int) -> tinfo_t:
        res = tif.get_member_recursive(typ, offset)
        return tinfo_t(res[1].type) if res is not None else VOID_POINTER_TYPE

    def _build_call_for_callback(self, ins: minsn_t, indirect_info: IndirectCallInfo | None) -> Call:
        """Build a Call object for the given instruction."""
        params: list[mcallarg_t] = list(ins.d.f.args)
        parsed_params = [_parse_param(param) for param in params]
        parsed_params_name = [_parse_param_name(param) for param in params]
        assignee = _try_extract_assignee(self.parents)
        return Call(self.mba.entry_ea, ins.ea, params, parsed_params, parsed_params_name, assignee, indirect_info)


def _try_extract_assignee(parents: list[mop_t | minsn_t]) -> mop_t | None:
    """
    Try to extract the assignee from the parents of the call.
    """
    if not parents:
        # Direct call instruction, no assignee
        return None
    assert len(parents) >= 2, f"Expected at least 2 parents, got {len(parents)}: {[p.dstr() for p in parents]}"

    # Skip the first parent, which is the mop_d wrapper
    parent = parents[-2]
    # Mop can be a child of the following:
    # - direct instruction operand (l,r,d)
    # - function argument
    # - ref to a variable - not if we are a function call
    # - pair - never seen this so let's ignore it
    if isinstance(parent, mop_t):
        # In this case we are a function argument, therefore no assignee
        assert parent.t == ida_hexrays.mop_f, f"Expected mop_f, got {parent.dstr()}"
        # If this is a dynamic cast, we can try to unwrap it
        call: minsn_t = parents[-3]
        if call.l.t == ida_hexrays.mop_h and any(
            call.l.helper.startswith(cast_name) for cast_name in CAST_FUNCTION_NAMES
        ):
            # Skip the cast helper and try to extract the assignee from the next parent
            return _try_extract_assignee(parents[:-3])

        return None
    # minsn_t case
    elif parent.is_like_move() or parent.opcode == ida_hexrays.m_stx:
        # Check there is an actual assignee and not a nop one
        if parent.d.t != ida_hexrays.mop_z:
            return parent.d

    # Not a move or store, no assignee
    return None


def _parse_param(param: mcallarg_t) -> ParsedParam:
    """Try to parse the param to python constant"""
    if _is_string_param(param):
        return mop.get_str(param)
    return mop.get_const_int(param)


def _parse_param_name(param: mcallarg_t) -> str | None:
    """Try to parse the parameter global name"""
    if param.t == ida_hexrays.mop_v:
        return memory.name_from_ea(param.g)
    elif param.t == ida_hexrays.mop_a and param.a.t == ida_hexrays.mop_v:
        return memory.name_from_ea(param.a.g)

    return None


def _is_string_param(param: mcallarg_t) -> bool:
    """Check if the param is a string pointer."""
    if param.type.is_ptr_or_array() and param.type.get_pointed_object().is_char():
        return True

    # Sometimes IDA decides it is a _QWORD even though the signature has char* and this is string const.
    # Let's assume it is a string param and see if it decodes
    return mop.get_str(param) is not None



__all__ = ["mark_all_outline_functions", "mark_outline_functions_component"]

import ida_kernwin
import idaapi
from ida_kernwin import action_handler_t

from ioshelper.base.reloadable_plugin import UIAction, UIActionsComponent

from .outline import mark_all_outline_functions

ACTION_ID = "ioshelper:mark_outline_components"

mark_outline_functions_component = UIActionsComponent.factory(
    "Locate all the outline functions and mark them as such",
    [
        lambda core: UIAction(
            ACTION_ID,
            idaapi.action_desc_t(
                ACTION_ID,
                "Locate all the outline functions and mark them as such",
                MarkAllOutlineFunctionsAction(),
            ),
            menu_location=UIAction.base_location(core),
        )
    ],
)


class MarkAllOutlineFunctionsAction(action_handler_t):
    def activate(self, ctx: ida_kernwin.action_ctx_base_t):
        mark_all_outline_functions()
        return 0

    def update(self, ctx):
        return idaapi.AST_ENABLE_ALWAYS



__all__ = ["os_log_enabled_optimizer_t"]

import re

import ida_hexrays
from ida_hexrays import (
    mblock_t,
    mcallinfo_t,
    minsn_t,
    mop_t,
    mop_visitor_t,
)
from idahelper.microcode import minsn

from ioshelper.base.utils import CounterMixin, match

from . import os_log

OSLOG_FUNCTIONS_TO_REPLACE_WITH_HELPER: list[str | re.Pattern] = [
    "_os_log_type_enabled",
    re.compile(r"_os_log_type_enabled_(\d+)"),
]
OSLOG_TYPE_INDEX = 1

SIGNPOST_FUNCTIONS_TO_REPLACE_WITH_HELPER: list[str | re.Pattern] = [
    "_os_signpost_enabled",
    re.compile(r"_os_signpost_enabled_(\d+)"),
]


class mop_optimizer_t(mop_visitor_t, CounterMixin):
    def visit_mop(self, op: mop_t, tp, is_target: bool) -> int:
        # No assignment dest, we want a call instruction
        if not is_target and op.d is not None:
            self.visit_instruction_mop(op)
        return 0

    def visit_instruction_mop(self, op: mop_t) -> None:
        # We only want calls
        insn: minsn_t = op.d
        if insn.opcode != ida_hexrays.m_call:
            return

        # Calls with names
        name = minsn.get_func_name_of_call(insn)
        if name is None:
            return

        # If it should be optimized to first arg, optimize
        if match(OSLOG_FUNCTIONS_TO_REPLACE_WITH_HELPER, name):
            fi: mcallinfo_t = insn.d.f
            if fi.args.empty():
                # No arguments, probably IDA have not optimized it yet
                return

            # Log type
            log_type_arg = fi.args[OSLOG_TYPE_INDEX]
            if log_type_arg.t != ida_hexrays.mop_n:
                return
            log_type = log_type_arg.unsigned_value()

            # Change name
            insn.l.make_helper(f"oslog_{os_log.log_type_to_str(log_type, is_signpost=False)}_enabled")
            self.count()
            # Remove arguments
            fi.args.clear()
            fi.solid_args = 0
            self.count()
        elif match(SIGNPOST_FUNCTIONS_TO_REPLACE_WITH_HELPER, name):
            fi: mcallinfo_t = insn.d.f

            # Change name
            insn.l.make_helper("ossignpost_enabled")
            self.count()
            # Remove arguments
            fi.args.clear()
            fi.solid_args = 0
            self.count()


class os_log_enabled_optimizer_t(ida_hexrays.optinsn_t):
    def func(self, blk: mblock_t, ins: minsn_t, optflags: int) -> int:
        # Let IDA reconstruct the calls before
        if blk.mba.maturity < ida_hexrays.MMAT_CALLS:
            return 0

        mop_optimizer = mop_optimizer_t(blk.mba, blk)
        ins.for_all_ops(mop_optimizer)
        changes = mop_optimizer.cnt
        if changes:
            blk.mark_lists_dirty()
        return changes



__all__ = ["mark_all_outline_functions"]


import re
import zlib
from collections import Counter
from itertools import islice

import ida_ua
from ida_funcs import func_t
from ida_ua import insn_t
from idahelper import functions, instructions, memory, xrefs

MAX_INSN_COUNT = 15
OUTLINE_COMMON_REGISTERS = ["X19", "X20", "X21", "X22", "X23", "X24", "X25"]
HASHES = {
    2486657537,
    4051340011,
    835854959,
    4053891209,
    2284322540,
    3734446324,
    3486938057,
    3100884732,
    1520871260,
    3035036398,
    3354748321,
    1286661669,
    3963943914,
    1065801208,
    3035825621,
    3374831749,
    3277914534,
    2587032924,
    2868248941,
    3315843181,
    2622615498,
    3720341170,
    3337984774,
    1893466989,
    518987256,
    2803156496,
    671639713,
    218684989,
    1791333242,
    3457579667,
    3574749379,
    3434086350,
    4165857220,
    2131640558,
    3495698196,
    1340405478,
    229647317,
    4006608232,
    786104779,
    1126219089,
    2831582611,
    289234555,
    1661182837,
    3423629888,
    2309108340,
    3075689215,
    2236221560,
    1839174898,
    481463652,
    766791884,
    218698196,
    752795631,
    3344164986,
    1827242795,
    792695013,
    230959901,
    2997841865,
    568754444,
    1670421527,
    2184318906,
    3024877629,
    2387594438,
    557192575,
    2458769492,
    3922284957,
    4107181330,
    1578820194,
    1546391466,
    1163279958,
    1540647309,
    1049262605,
    2130556960,
    2163533787,
    2092533000,
    2426425641,
    4207618,
    2736070239,
    1277505515,
    1788212028,
    1368299300,
    3786044294,
    563839748,
    3199076566,
    876720633,
    2633791921,
    1428282898,
    146181378,
    472218750,
    304360983,
    1808920834,
    4174859059,
    85578290,
    631075697,
    1189335917,
    288603166,
    4177653000,
    3698829753,
    1630600770,
    1498027780,
    1149213890,
}


def calculate_outline_hash(top_hashes_count: int):
    # Calculate hashes
    hashes = Counter()
    for func in functions.iterate_functions():
        func_name = memory.name_from_ea(func.start_ea)
        assert func_name is not None

        if functions.has_flags(func, functions.FLAG_OUTLINE):
            hashes[function_hash(func)] += 1
            continue

    top_hashes = [key for key, _ in hashes.most_common(top_hashes_count)]
    top_hashes_match = sum(value for _, value in hashes.most_common(top_hashes_count))
    print(f"Top {top_hashes_count} hashes, should match {top_hashes_match} / {hashes.total()}")
    print(top_hashes)

    # Check for false positives
    top_hashes = set(top_hashes)
    for func in functions.iterate_functions():
        func_name = memory.name_from_ea(func.start_ea)
        assert func_name is not None

        if not functions.has_flags(func, functions.FLAG_OUTLINE):
            func_hash = function_hash(func)
            if func_hash in top_hashes:
                print(f"{func.start_ea:X} matches an hash for outlined func.")


def mark_all_outline_functions():
    count = 0
    for func in functions.iterate_functions():
        func_name = memory.name_from_ea(func.start_ea)
        assert func_name is not None

        # if functions.has_flags(func, functions.FLAG_OUTLINE):
        #     continue
        if func_name.startswith("_OUTLINED") or (
            not heuristic_not_outline(func, func_name) and (function_hash(func) in HASHES or heuristic_outline(func))
        ):
            print(f"Applied outline flag on {func.start_ea:X}")
            functions.apply_flag_to_function(func, functions.FLAG_OUTLINE)

            # Update name as well
            if "OUTLINED" not in func_name:
                memory.set_name(func.start_ea, f"__OUTLINED_{func_name}")
            count += 1

    print(f"Applied outlined to {count} functions")


def heuristic_not_outline(func: func_t, name: str) -> bool:
    # Assuming outlined functions are small - less than 10 instructions.
    if func.size() > MAX_INSN_COUNT * 4:
        return True

    if not name.startswith("sub_"):
        return True

    # Outlined functions will have no data xrefs
    if xrefs.get_xrefs_to(func.start_ea, is_data=True):
        return True

    # We only care for functions with xrefs.
    # One might think outline functions has to have more than one xref, but apparently it is incorrect...
    if not xrefs.get_xrefs_to(func.start_ea):
        return True

    first_instruction = instructions.decode_instruction(func.start_ea)
    return bool(first_instruction is None or first_instruction.get_canon_mnem() in ["PAC", "BTI"])


def heuristic_outline(func: func_t) -> bool:
    # Empty function are not outlined I guess
    if func.size() == 0:
        return False

    first_insn = instructions.decode_instruction(func.start_ea)
    if first_insn is None:
        return False

    # We don't deal right now with instructions that may branch on first instruction
    if instructions.is_flow_instruction(first_insn):
        return False

    # Check if we use an outline register without definition in the first instruction
    read_0, write_0 = instructions.analyze_reg_dependencies(first_insn)
    if any(reg in read_0 for reg in OUTLINE_COMMON_REGISTERS):
        return True

    # Check if we use an outline register without definition in the second instruction
    second_insn = instructions.decode_next_instruction(first_insn, func)
    if second_insn is None:
        return False

    read_1, _ = instructions.analyze_reg_dependencies(second_insn)
    return any(reg in read_1 and reg not in write_0 for reg in OUTLINE_COMMON_REGISTERS)


REG_GROUPS = {
    "a": range(0, 8),  # x0-x7
    "t": range(8, 16),  # x8-x15
    "i": range(16, 19),  # x16-x18
    "s1": range(19, 24),  # x19-x23
    "s2": range(24, 29),  # x24-x28
    "fp": [29],
    "lr": [30],
    "z": ["xzr"],
}


def classify_reg(reg: int) -> str:
    reg_name = instructions.get_register_name(reg).lower().strip()
    if reg_name == "sp":
        return "sp"
    if reg_name == "xzr" or reg_name == "wzr":
        return "z"
    if reg_name.startswith(("x", "w")):
        regnum = int(re.findall(r"\d+", reg_name)[0])
        for group, rset in REG_GROUPS.items():
            if regnum in rset:
                return group
        return "r"  # everything else
    return reg_name  # leave non-registers as-is


def get_normalized_pattern(insn: insn_t) -> str:
    mnem = insn.get_canon_mnem()
    ops = []
    for op in insn.ops:
        if op.type == ida_ua.o_void:
            reg = "V"
        elif op.type == ida_ua.o_imm:
            reg = str(op.value)
        elif op.type == ida_ua.o_mem:
            reg = "M"
        elif op.type == ida_ua.o_reg:
            reg = classify_reg(op.reg)
        elif op.type == ida_ua.o_phrase or op.type == ida_ua.o_idpspec0:
            reg = f"{classify_reg(op.reg)}:{op.phrase}"
        elif op.type == ida_ua.o_displ:
            reg = f"{classify_reg(op.reg)}:{op.phrase}:{op.value}"
        else:
            reg = f"U{op.type}"

        ops.append(reg)
    return f"{mnem}_{'_'.join(ops)}"


def get_function_pattern(func) -> str:
    patterns = []
    for insn in islice(instructions.from_func(func), 2):
        patterns.append(get_normalized_pattern(insn))
    return " ".join(patterns)


def function_hash(func) -> int:
    pattern = get_function_pattern(func)
    return zlib.crc32(pattern.encode())



__all__ = ["log_error_case_optimizer_t"]

import re

import ida_hexrays
from ida_hexrays import mblock_t, minsn_t, minsn_visitor_t, mop_t, mop_visitor_t, optinsn_t
from idahelper.microcode import mop

from ioshelper.base.utils import CounterMixin, match_dict

# Replace var with val
VARIABLES_TO_OPTIMIZE_OUT: dict[str | re.Pattern, int] = {
    "_gNumLogObjects": 0x1000,
    re.compile(r"_gNumLogObjects_(\d+)"): 0x1000,
    "_gNumLogSignpostObjects": 0x1000,
    re.compile(r"_gNumLogSignpostObjects_(\d+)"): 0x1000,
}

# replace jz var, 0, #addr with goto/nop
# bool is "isZero"
JZ_TO_OPTIMIZE: dict[str | re.Pattern, bool] = {
    "_gLogObjects": False,
    re.compile(r"_gLogObjects_(\d+)"): False,
    "_gLogSignpostObjects": False,
    re.compile(r"_gLogSignpostObjects_(\d+)"): False,
}


class variable_optimizer_t(mop_visitor_t, CounterMixin):
    def visit_mop(self, op: mop_t, tp, is_target) -> int:
        # we want global mops
        if is_target or op.g is None:
            return 0

        # We want named global
        name = mop.get_name(op)
        if name is None:
            return 0

        # Convert number to the optimized value
        if (val := match_dict(VARIABLES_TO_OPTIMIZE_OUT, name)) is not None:
            op.make_number(val, op.size)
            self.count()

        return 0


class jz_optimizer_t(minsn_visitor_t, CounterMixin):
    def visit_minsn(self) -> int:
        insn: minsn_t = self.curins
        if insn.opcode not in [ida_hexrays.m_jnz, ida_hexrays.m_jz]:
            return 0

        # We want conditions on global variables
        name = mop.get_name(insn.l)
        if name is None:
            return 0

        if (is_zero := match_dict(JZ_TO_OPTIMIZE, name)) is not None:
            #     not zero, zero
            # jnz     1       0
            # jz      0       1
            should_jmp = is_zero == (insn.opcode == ida_hexrays.m_jz)

            # We don't optimize it directly to goto/nop, since it will require blocks in/out.
            # IDA can optimize it for us :)
            insn.l.make_number(0, 1)
            insn.r.make_number(0, 1)
            insn.opcode = ida_hexrays.m_jz if should_jmp else ida_hexrays.m_jnz
            self.count()

        return 0


class log_error_case_optimizer_t(optinsn_t):
    def func(self, blk: mblock_t, insn: minsn_t, optflags: int) -> int:
        if blk.mba.maturity < ida_hexrays.MMAT_CALLS:
            return 0

        variable_optimizer = variable_optimizer_t()
        jz_optimizer = jz_optimizer_t()
        insn.for_all_ops(variable_optimizer)
        insn.for_all_insns(jz_optimizer)
        cnt = variable_optimizer.cnt + jz_optimizer.cnt

        if cnt:
            blk.mark_lists_dirty()
        return cnt



# This code was inspired by Lucid's plugin core, which is licensed under the MIT License.
# However, this is a complete rewrite that only contains the _startup_hooks part from the original code.
# As such, I find it acceptable to use this code under the same license as the rest of the plugin.
# https://github.com/gaasedelen/lucid/blob/master/plugins/lucid/core.py
import abc
import contextlib
import dataclasses
import sys
from collections.abc import Callable
from pathlib import Path
from typing import Protocol

import ida_diskio
import ida_hexrays
import ida_idaapi
import ida_kernwin
import idaapi
from ida_idaapi import plugin_t
from ida_kernwin import UI_Hooks, action_desc_t

IS_DEBUG = (Path(__file__).parent.parent.parent / "DEBUG").exists()

DISABLED_PLUGINS_PATH = Path(ida_diskio.get_user_idadir()) / "ioshelper_disabled_plugins.txt"


def get_disabled_plugins() -> set[str]:
    """Return a list of disabled plugins."""
    if not DISABLED_PLUGINS_PATH.exists():
        return set()

    disabled_plugins: set[str] = set()
    with DISABLED_PLUGINS_PATH.open("r") as f:
        for line in f:
            line = line.strip()
            if line and not line.startswith("#"):
                disabled_plugins.add(line)
    return disabled_plugins


class Component:
    """
    A component is a self-contained piece of functionality that can be loaded and unloaded independently.
    It will only be loaded and unloaded when the plugin core is loaded and unloaded.
    However, it can be mounted and unmounted independently of the plugin core.
    """

    def __init__(self, name: str, core: "PluginCore"):
        self.name = name
        self.core = core

    def load(self) -> bool:
        """Load the component and all the relevant resources"""
        return True

    def mount(self) -> bool:
        """Enable the functionality of the component"""
        return True

    def unmount(self):
        """Disable the functionality of the component"""
        pass

    def unload(self):
        """Unload the component and all the relevant resources"""
        pass


ComponentFactory = Callable[["PluginCore"], Component]
RunCallbackFactory = Callable[["PluginCore"], Callable[[int], None]]


class PluginCoreFactory(Protocol):
    def __call__(self, defer_load: bool, should_mount: bool) -> float: ...


class PluginCore:
    def __init__(
        self,
        name: str,
        component_factories: list[ComponentFactory],
        run_callback_factory: RunCallbackFactory,
        defer_load: bool = False,
        should_mount: bool = True,
    ):
        self.name = name
        self.run_callback = run_callback_factory(self)
        self.loaded = False
        self.mounted = False
        all_components = [factory(self) for factory in component_factories]
        self._components = [component for component in all_components if component.name not in get_disabled_plugins()]

        # we can 'defer' the load of the plugin core a little bit. this
        # ensures that all the other plugins (eg, decompilers) can get loaded
        # and initialized when opening an idb/bin

        def perform_load():
            self.load()
            if should_mount:
                self.mount()

        class UIHooks(UI_Hooks):
            def ready_to_run(self):
                perform_load()

        self._startup_hooks = UIHooks()

        if defer_load:
            self._startup_hooks.hook()
        else:
            perform_load()

    def load(self):
        self._startup_hooks.unhook()

        if not ida_hexrays.init_hexrays_plugin():
            print(f"[{self.name}] failed to load hexrays plugin, aborting load.")
            return

        print(f"[{self.name}] loading plugin")
        for component in self._components:
            print(f"[{self.name}] loading component {component.name}")
            if not component.load():
                print(f"[{self.name}] failed to load component {component.name}, aborting load.")
                return

        self.loaded = True

    def mount(self):
        if not self.mounted:
            for component in self._components:
                if self.should_mount(component):
                    print(f"[{self.name}] mounting component {component.name}")
                    if not component.mount():
                        print(f"[{self.name}] failed to mount component {component.name}, aborting mount.")
                        return
            self.mounted = True

    def unmount(self):
        if self.mounted:
            for component in self._components:
                print(f"[{self.name}] unmounting component {component.name}")
                component.unmount()

            self.mounted = False

    def unload(self):
        """
        Unload the plugin core.
        """

        # unhook just in-case load() was never actually called...
        self._startup_hooks.unhook()

        # if the core was never fully loaded, there's nothing else to do
        if not self.loaded:
            return

        print(f"[{self.name}] unloading plugin")

        # mark the core as 'unloaded' and teardown its components
        self.loaded = False

        self.unmount()
        for component in self._components:
            print(f"[{self.name}] unloading component {component.name}")
            component.unload()

    def should_mount(self, _component: Component) -> bool:
        """
        Determine if a component should be mounted based on the current state of the plugin core.
        In the future, we will implement a more sophisticated system to determine if a component should be mounted.
        """
        return True

    def run(self, arg: int):
        """
        Proxy for the `run` method of the plugin_t interface.
        """
        self.run_callback(arg)

    @staticmethod
    def factory(
        name: str, component_factories: list[ComponentFactory], run_callback_factory: RunCallbackFactory
    ) -> PluginCoreFactory:
        def plugin_core_factory(defer_load: bool, should_mount: bool) -> PluginCore:
            return PluginCore(
                name, component_factories, run_callback_factory, defer_load=defer_load, should_mount=should_mount
            )

        # The type checker seems to have trouble with the factory method, so we need to suppress it
        # noinspection PyTypeChecker
        return plugin_core_factory


class ReloadablePlugin(abc.ABC, plugin_t):
    def __init__(
        self,
        global_name: str,
        base_package_name: str,
        plugin_core_factory: PluginCoreFactory,
        extra_packages_to_reload: list[str] | None = None,
    ):
        super().__init__()
        self._global_name = global_name
        self._plugin_core_factory = plugin_core_factory
        self._base_package_name = base_package_name
        self.core: PluginCore | None = None
        self._reload_plugin_action_id = f"{self._global_name}:reload_plugin"
        self.extra_packages_to_reload = extra_packages_to_reload or []

    def init(self) -> int:
        self.core = self._plugin_core_factory(defer_load=True, should_mount=True)
        # Provide access from ida python console
        setattr(sys.modules["__main__"], self._global_name, self)
        # register the reload action
        idaapi.register_action(
            idaapi.action_desc_t(
                self._reload_plugin_action_id,
                f"Reload plugin: {getattr(self, 'wanted_name', self._global_name)}",
                PluginReloadActionHandler(self),
                "f2" if IS_DEBUG else None,
            )
        )
        # Keep plugin alive
        return ida_idaapi.PLUGIN_KEEP

    def term(self) -> None:
        idaapi.unregister_action(self._reload_plugin_action_id)
        if self.core is not None:
            self.core.unload()

    def run(self, arg: int):
        if self.core is not None:
            self.core.run(arg)

    def reload(self):
        """Hot-reload the plugin core."""
        print(f"[{getattr(self, 'wanted_name', 'plugin')}] Reloading...")

        # Unload the core and all its components
        was_mounted = self.core.mounted if self.core else True
        if self.core is not None:
            self.core.unload()

        # Reload all modules in the base package
        modules_to_reload = [module_name for module_name in sys.modules if self.should_reload_pkg(module_name)]
        for module_name in modules_to_reload:
            with contextlib.suppress(ModuleNotFoundError):
                idaapi.require(module_name)

        # Load the plugin core
        self.core = self._plugin_core_factory(defer_load=False, should_mount=was_mounted)

    def should_reload_pkg(self, module_name: str) -> bool:
        """Should we reload this module on reloading the plugin"""
        if module_name.startswith(self._base_package_name):
            return True
        return any(module_name.startswith(prefix) for prefix in self.extra_packages_to_reload)


class PluginReloadActionHandler(ida_kernwin.action_handler_t):
    def __init__(self, plugin: ReloadablePlugin):
        super().__init__()
        self.plugin = plugin

    def activate(self, ctx):
        self.plugin.reload()
        print("Reloaded plugin!")
        return 1

    def update(self, ctx) -> int:
        return idaapi.AST_ENABLE_ALWAYS


# A common type of component is installing optimizers for the decompiler. This is a helper class to make it easier.

optimizer_t = ida_hexrays.optblock_t | ida_hexrays.optinsn_t
optimizer_factory_t = Callable[[], optimizer_t]


class OptimizersComponent(Component):
    def __init__(self, name: str, core: PluginCore, optimizer_factories: list[optimizer_factory_t]):
        super().__init__(name, core)
        self._optimizer_factories = optimizer_factories
        self._optimizers: list[optimizer_t] | None = None

    def load(self) -> bool:
        self._optimizers = [factory() for factory in self._optimizer_factories]
        return True

    def mount(self) -> bool:
        assert self._optimizers is not None, "Load must be called before mount"

        for optimizer in self._optimizers:
            optimizer.install()
        return True

    def unmount(self):
        assert self._optimizers is not None, "Load must be called before unmount"

        for optimizer in self._optimizers:
            optimizer.remove()

    def unload(self):
        self._optimizers = None

    @staticmethod
    def factory(name: str, optimizer_factories: list[optimizer_factory_t]) -> ComponentFactory:
        return lambda core: OptimizersComponent(name, core, optimizer_factories)


@dataclasses.dataclass
class UIAction:
    id: str
    action_desc: action_desc_t
    menu_location: str | None = None
    dynamic_menu_add: Callable[["TWidget *", "TPopupMenu *"], bool] | None = None  # noqa: F722

    @staticmethod
    def base_location(core: PluginCore) -> str:
        """
        Returns the base location for the plugin's UI actions.
        This is used to create a unique menu location for the actions.
        """
        return f"Edit/Plugins/{core.name}/"


class UIActionsComponentUIHooks(idaapi.UI_Hooks):
    def __init__(self, actions: list[UIAction]):
        super().__init__()
        self._actions = actions

    def finish_populating_widget_popup(self, widget, popup):
        for action in self._actions:
            if action.dynamic_menu_add is not None and action.dynamic_menu_add(widget, popup):
                idaapi.attach_action_to_popup(widget, popup, action.id)


# Another common type of component is installing ui actions. This is a helper class to make it easier.
class UIActionsComponent(Component):
    def __init__(self, name: str, core: PluginCore, action_factories: list[Callable[[PluginCore], UIAction]]):
        super().__init__(name, core)
        self._action_factories = action_factories
        self._actions: list[UIAction] | None = None
        self._ui_hooks: UI_Hooks | None = None

    def load(self) -> bool:
        self._actions = [factory(self.core) for factory in self._action_factories]
        if any(action.dynamic_menu_add is not None for action in self._actions):
            # Create a UI_Hooks instance to attach the dynamic menu
            self._ui_hooks = UIActionsComponentUIHooks(self._actions)
            self._ui_hooks.hook()

        for action in self._actions:
            if not idaapi.register_action(action.action_desc):
                print(f"[{self.name}] failed to register action {action.id}, aborting load.")
                return False

            if action.menu_location is not None and not idaapi.attach_action_to_menu(
                action.menu_location,
                action.id,
                0,
            ):
                print(
                    f"[{self.name}] failed to attach action {action.id} to menu {action.menu_location}, aborting load."
                )
                return False

        return True

    def unload(self):
        for action in self._actions:
            idaapi.unregister_action(action.id)

        if self._ui_hooks is not None:
            self._ui_hooks.unhook()

    @staticmethod
    def factory(name: str, action_factories: list[Callable[[PluginCore], UIAction]]) -> ComponentFactory:
        return lambda core: UIActionsComponent(name, core, action_factories)


# A common type of component is installing hooks for the decompiler. This is a helper class to make it easier.
class HexraysHookComponent(Component):
    def __init__(self, name: str, core: PluginCore, hook_factories: list[Callable[[], ida_hexrays.Hexrays_Hooks]]):
        super().__init__(name, core)
        self._hook_factories = hook_factories
        self._hooks: list[ida_hexrays.Hexrays_Hooks] | None = None

    def load(self):
        self._hooks = [hook_factory() for hook_factory in self._hook_factories]
        return True

    def mount(self) -> bool:
        assert self._hooks is not None, "load() must be called before mount()"
        for hook in self._hooks:
            hook.hook()
        return True

    def unmount(self):
        if self._hooks is not None:
            for hook in self._hooks:
                hook.unhook()

    def unload(self):
        self._hooks = None

    @staticmethod
    def factory(name: str, hook_factories: list[Callable[[], ida_hexrays.Hexrays_Hooks]]) -> ComponentFactory:
        return lambda core: HexraysHookComponent(name, core, hook_factories)



__all__ = ["plugin_core"]

import ida_kernwin
import idaapi
from idahelper import file_format, widgets

from .base.reloadable_plugin import ComponentFactory, PluginCore, UIAction, UIActionsComponent
from .plugins.common.clang_blocks import clang_block_args_analyzer_component, clang_block_optimizer_component
from .plugins.common.jump_to_string import jump_to_string_component
from .plugins.common.outline import mark_outline_functions_component
from .plugins.common.range_condition import range_condition_optimizer_component
from .plugins.common.run_callback import run_callback
from .plugins.common.segment_xrefs import show_segment_xrefs_component
from .plugins.kernelcache.cpp_vtbl import jump_to_vtable_component
from .plugins.kernelcache.func_renamers import (
    apply_pac_component,
    local_func_renamer_component,
    mass_func_renamer_component,
)
from .plugins.kernelcache.generic_calls_fix import generic_calls_fix_component
from .plugins.kernelcache.kalloc_type import apply_kalloc_type_component
from .plugins.kernelcache.obj_this import this_arg_fixer_component
from .plugins.objc.objc_ref import objc_xrefs_component
from .plugins.objc.objc_refcnt import component as objc_refcount_component
from .plugins.objc.objc_sugar import objc_sugar_component
from .plugins.objc.oslog import component as oslog_component

TOGGLE_ACTION_ID = "ioshelper:toggle"

toggle_ios_helper_mount_component = UIActionsComponent.factory(
    "toggle plugin mounting",
    [
        lambda core: UIAction(
            TOGGLE_ACTION_ID,
            idaapi.action_desc_t(
                TOGGLE_ACTION_ID,
                "Toggle iOS helper optimizations",
                IOSHelperToggleActionHandler(core),
            ),
            menu_location=UIAction.base_location(core),
        )
    ],
)


class IOSHelperToggleActionHandler(ida_kernwin.action_handler_t):
    def __init__(self, core: PluginCore):
        super().__init__()
        self.core = core

    def activate(self, ctx):
        if self.core.mounted:
            self.core.unmount()
        else:
            self.core.mount()

        widgets.refresh_pseudocode_widgets()

        print("Obj-C optimization are now:", "enabled" if self.core.mounted else "disabled")
        print("Note: You might need to perform decompile again for this change to take effect.")
        return 1

    def update(self, ctx) -> int:
        return idaapi.AST_ENABLE_ALWAYS


def get_modules_for_file() -> list[ComponentFactory]:
    return [
        *shared_modules(),
        *(objc_plugins() if file_format.is_objc() else []),
        *(kernel_cache_plugins() if file_format.is_kernelcache() else []),
    ]


def shared_modules() -> list[ComponentFactory]:
    return [
        toggle_ios_helper_mount_component,
        clang_block_args_analyzer_component,
        clang_block_optimizer_component,
        jump_to_string_component,
        objc_refcount_component,
        range_condition_optimizer_component,
        mark_outline_functions_component,
        show_segment_xrefs_component,
    ]


def objc_plugins() -> list[ComponentFactory]:
    return [
        oslog_component,
        objc_xrefs_component,
        objc_sugar_component,
    ]


def kernel_cache_plugins() -> list[ComponentFactory]:
    return [
        this_arg_fixer_component,
        jump_to_vtable_component,
        generic_calls_fix_component,
        local_func_renamer_component,
        mass_func_renamer_component,
        apply_kalloc_type_component,
        apply_pac_component,
    ]


plugin_core = PluginCore.factory("iOSHelper", get_modules_for_file(), run_callback)



import ida_hexrays
from ida_hexrays import cexpr_t
from ida_typeinf import tinfo_t
from idahelper import cpp, memory, tif, widgets
from idahelper.widgets import EAChoose


def get_vtable_call(verbose: bool = False) -> tuple[tinfo_t, str, int] | None:
    """If the mouse is on a virtual call, return the vtable type, method name and offset."""
    citem = widgets.get_current_citem()
    if citem is None:
        if verbose:
            print("[Error] No citem found. Do you have your cursor on a virtual call?")
        return None
    if not citem.is_expr():
        if verbose:
            print(
                f"[Error] Current citem is not an expression: {citem.dstr()}. Do you have your cursor on the virtual call?"
            )
        return None

    return get_vtable_call_from_expr(citem.cexpr, verbose)


def get_vtable_call_from_expr(expr: cexpr_t, verbose: bool = False) -> tuple[tinfo_t, str, int] | None:
    if expr.op not in [ida_hexrays.cot_memptr, ida_hexrays.cot_memref]:
        if verbose:
            print(
                f"[Error] Current citem is not a member pointer: {expr.dstr()} but a {ida_hexrays.get_ctype_name(expr.op)}. Do you have your cursor on the virtual call?"
            )
        return None

    tp: tinfo_t = expr.type
    if not tp.is_funcptr() and not tp.is_func():
        if verbose:
            print(
                f"[Error] Current member is not a function pointer: {expr.dstr()}. Do you have your cursor on a virtual call?"
            )
        return None
    offset = expr.m
    vtable_type = expr.x.type

    # A bit hack but should work. We could implement a better way to get the name in the future...
    call_name = expr.dstr().split(".")[-1].split("->")[-1]
    return vtable_type, call_name, offset


def show_vtable_xrefs():
    vtable_call = get_vtable_call(verbose=True)
    if vtable_call is None:
        return

    vtable_type, call_name, offset = vtable_call
    actual_type = get_actual_class_from_vtable(vtable_type)
    if actual_type is None:
        print(f"[Error] failed to find actual type for {vtable_type.get_type_name()}")
        return

    matches = get_vtable_xrefs(vtable_type, offset)

    method_name = f"{actual_type.get_type_name()}->{call_name}"
    if not matches:
        print(f"[Error] No implementations found for {method_name}")
    if len(matches) == 1:
        # Just jump to the function
        widgets.jump_to(next(iter(matches.keys())))
    elif matches:
        # Show the results in a chooser
        print(f"Implementations for {method_name}:")
        for ea, cls in matches.items():
            print(f"{hex(ea)}: {memory.name_from_ea(ea)} by {cls}")

        xrefs_choose = EAChoose(
            f"Implementations for {method_name}",
            list(matches.items()),
            col_names=("EA", "Implementing class"),
            modal=True,
        )
        xrefs_choose.show()


def get_vtable_xrefs(vtable_type: tinfo_t, offset: int) -> dict[int, str]:
    """Given a vtable type and offset, return the address of the function at that offset."""
    actual_type = get_actual_class_from_vtable(vtable_type)
    if actual_type is None:
        return {}

    children_classes = tif.get_children_classes(actual_type) or []
    pure_virtual_ea = (
        memory.ea_from_name("___cxa_pure_virtual")
        or memory.ea_from_name("__cxa_pure_virtual")
        or memory.ea_from_name("_cxa_pure_virtual")
    )
    assert pure_virtual_ea is not None
    matches: dict[int, str] = {}  # addr -> class_name

    # Get the base implementation, either from this class or its parent if it is inherited
    parent_impl = get_impl_from_parent(actual_type, offset, pure_virtual_ea)
    if parent_impl is not None:
        matches[parent_impl[0]] = parent_impl[1]

    for cls in children_classes:
        vtable_func_ea = get_vtable_entry(cls, offset, pure_virtual_ea)
        if vtable_func_ea is None:
            continue

        # Add it to the dict if not already present.
        # get_children_classes returns the classes in order of inheritance
        if vtable_func_ea not in matches:
            # noinspection PyTypeChecker
            matches[vtable_func_ea] = cls.get_type_name()  # pyright: ignore[reportArgumentType]
    return matches


def get_impl_from_parent(cls: tinfo_t, offset: int, pure_virtual_ea: int) -> tuple[int, str] | None:
    """
    Given a class and an offset to vtable entry, Iterate over its parents to find what will be the implementation
    for the given offset. If no implementation is found, return None.
    """
    impl, impl_cls = get_vtable_entry(cls, offset, pure_virtual_ea), cls.get_type_name()
    if impl is None:
        # If not implemented in this class, will not be implemented in its parents.
        return None
    for parent_cls in tif.get_parent_classes(cls):
        if offset >= cpp.vtable_methods_count(parent_cls, False) * 8:
            # If offset is greater than the size of the vtable, the method was defined in child class
            break
        this_impl = get_vtable_entry(parent_cls, offset, pure_virtual_ea)
        if this_impl is None or impl != this_impl:
            break
        else:
            impl_cls = parent_cls.get_type_name()

    return impl, f"{impl_cls} (Slot at {cls.get_type_name()})"


def get_vtable_entry(cls: tinfo_t, offset: int, pure_virtual_ea: int) -> int | None:
    """Given a class and an offset to vtable entry, return the ea of the function at the given offset
    if it is not pure virtual."""
    vtable_func_ea = cpp.vtable_func_at(cls, offset)
    return vtable_func_ea if vtable_func_ea and pure_virtual_ea != vtable_func_ea else None


def get_actual_class_from_vtable(vtable_type: tinfo_t) -> tinfo_t | None:
    # It is usually a pointer to a pointer to a vtable
    if vtable_type.is_ptr():
        vtable_type = vtable_type.get_pointed_object()

    return tif.type_from_vtable_type(vtable_type)



__all__ = ["can_show_segment_xrefs", "get_current_expr", "show_segment_xrefs"]

from dataclasses import dataclass

import ida_hexrays
import idaapi
from ida_hexrays import cexpr_t, cfuncptr_t, cinsn_t, citem_t, get_widget_vdui
from ida_kernwin import Choose
from ida_typeinf import tinfo_t
from idahelper import cpp, functions, memory, segments, tif, widgets, xrefs


def show_segment_xrefs(expr: cexpr_t, func_ea: int):
    if not _can_show_segment_xrefs(expr):
        print(f"[Error] Cannot show segment xrefs for expression: {expr.dstr()}")
        return

    if expr.op == ida_hexrays.cot_obj:
        segment_xrefs = _ea_to_xrefs(expr.obj_ea)
        title = f"Segment Xrefs for {memory.name_from_ea(expr.obj_ea) or f'{expr.obj_ea:#X}'}"
    else:
        segment_xrefs = _expr_to_xrefs(expr, func_ea)
        title = f"Segment Xrefs for {expr.dstr()}"

    if not segment_xrefs:
        print("No segment xrefs found.")
        return

    res = XrefsChoose(title, segment_xrefs).show()
    if not res:
        print("[Error] failed to show segment xrefs.")
        return


@dataclass
class SegmentXref:
    address: int
    function: int
    line: str


def _ea_to_xrefs(ea: int) -> list[SegmentXref]:
    """Get all xrefs to the given EA"""
    segment_xrefs: list[SegmentXref] = []
    for xref in xrefs.code_xrefs_to(ea):
        decompiled = idaapi.decompile(ea, flags=ida_hexrays.DECOMP_GXREFS_FORCE)
        if decompiled is None:
            print(f"[Warning] Could not decompile function at {xref:X}")
            continue
        decompiled_line = _get_decompiled_line(decompiled, xref) or "<error>"
        segment_xrefs.append(SegmentXref(xref, decompiled.entry_ea, decompiled_line))

    return segment_xrefs


def _expr_to_xrefs(expr: cexpr_t, func_ea: int) -> list[SegmentXref]:
    assert expr.op in (ida_hexrays.cot_memptr, ida_hexrays.cot_memref)
    expr_obj_type: tinfo_t = expr.x.type
    if expr_obj_type.is_ptr_or_array():
        expr_obj_type.remove_ptr_or_array()
    expr_obj_offset: int = expr.m
    recursive_member = tif.get_member_recursive(expr_obj_type, expr_obj_offset)
    if recursive_member is None:
        print(f"[Error] Could not find member at offset {expr_obj_offset} in type {expr_obj_type.dstr()}")
        return []
    relevant_type: tinfo_t = recursive_member[0]
    possible_types: set[str] = {str(t) for t in tif.get_children_classes(relevant_type) or []}
    possible_types.add(str(relevant_type))

    current_segment = segments.get_segment_by_ea(func_ea)
    if current_segment is None:
        print(f"[Error] Could not find segment for function at {func_ea:X}")
        return []

    segment_xrefs: list[SegmentXref] = []
    for func in functions.iterate_functions(current_segment.start_ea, current_segment.end_ea):
        decompiled = idaapi.decompile(func.start_ea, flags=ida_hexrays.DECOMP_GXREFS_FORCE)
        if decompiled is None:
            print(f"[Warning] Could not decompile function at {func.start_ea:X}")
            continue

        segment_xrefs.extend(_find_xrefs_to_field(possible_types, expr_obj_offset, decompiled))

    return segment_xrefs


def _find_xrefs_to_field(possible_types: set[str], offset: int, func: cfuncptr_t) -> list[SegmentXref]:
    """Find all xrefs to the given field in the context of the function."""
    segment_xrefs: list[SegmentXref] = []
    for item in func.treeitems:
        actual_item: cexpr_t | cinsn_t = item.to_specific_type
        # Check if it is field access
        if not isinstance(actual_item, cexpr_t) or actual_item.op not in [
            ida_hexrays.cot_memptr,
            ida_hexrays.cot_memref,
        ]:
            continue
        # Check if the type and offset match
        item_type: tinfo_t = actual_item.x.type
        if item_type.is_ptr_or_array():
            item_type.remove_ptr_or_array()
        if str(item_type) not in possible_types or actual_item.m != offset:
            continue

        container_insn = _find_first_container_instruction(actual_item, func)
        if container_insn is None:
            print(f"[Warning] Could not find container instruction for item: {actual_item.dstr()}")
            continue

        item_ea = container_insn.ea
        segment_xrefs.append(SegmentXref(item_ea, func.entry_ea, container_insn.dstr()))
    return segment_xrefs


def _find_first_container_instruction(item: citem_t | None, func: cfuncptr_t) -> cinsn_t | None:
    """Find the EA of the given item in the context of the function."""
    while item is not None:
        if isinstance(item, cinsn_t):
            return item
        item = func.body.find_parent_of(item)
        if item is not None:
            item = item.to_specific_type

    return None


def _get_decompiled_line(func: cfuncptr_t, ea: int) -> str | None:
    """Get the decompiled line for the given EA in the context of the function."""
    ea_map = func.get_eamap()
    if ea not in ea_map:
        print(f"[Warning] {ea:X} is not in {func.entry_ea:X} ea map.")
        return None

    return "\n".join(stmt.dstr() for stmt in ea_map[ea])


def _can_show_segment_xrefs(expr: cexpr_t) -> bool:
    """Check if we can show segment xrefs for the given expression."""
    return expr.op in (ida_hexrays.cot_obj, ida_hexrays.cot_memref, ida_hexrays.cot_memptr)


def can_show_segment_xrefs(widget) -> bool:
    """Check if we can show segment xrefs in the current context."""
    expr = get_current_expr(widget)
    return expr is not None and _can_show_segment_xrefs(expr)


def get_current_expr(widget) -> cexpr_t | None:
    """Get the current expression in the context."""
    if idaapi.get_widget_type(widget) != idaapi.BWN_PSEUDOCODE:
        return None
    vu = get_widget_vdui(widget)
    if not vu or not vu.item or vu.item.citype != ida_hexrays.VDI_EXPR:
        return None
    return vu.item.it.to_specific_type


class XrefsChoose(Choose):
    def __init__(self, title: str, items: list[SegmentXref]):
        Choose.__init__(
            self,
            title,
            [
                ["Address", 20 | Choose.CHCOL_EA],
                ["Function", 40 | Choose.CHCOL_FNAME],
                ["Line", 40 | Choose.CHCOL_PLAIN],
            ],
            flags=Choose.CH_RESTORE,
            embedded=False,
        )
        self.items = items
        self.modal = False

    def OnInit(self) -> bool:
        return True

    def OnGetSize(self) -> int:
        return len(self.items)

    def OnGetLine(self, n):
        item = self.items[n]
        return (
            f"{item.address:X}",
            cpp.demangle_name_only(memory.name_from_ea(item.function) or f"SUB_{item.function:X}"),
            item.line,
        )

    def OnGetEA(self, n) -> int:
        return self.items[n].address

    def OnSelectLine(self, n):
        ea = self.items[n].address
        widgets.jump_to(ea)
        return (Choose.NOTHING_CHANGED,)

    def show(self) -> bool:
        ok = self.Show(self.modal) >= 0
        return ok



__all__ = ["this_arg_fixer_component"]

import ida_kernwin
import idaapi
from ida_kernwin import action_handler_t
from idahelper import widgets

from ioshelper.base.reloadable_plugin import UIAction, UIActionsComponent

from .obj_this import update_argument

ACTION_ID = "ioshelper:this_arg_fixer"

this_arg_fixer_component = UIActionsComponent.factory(
    "Convert first argument to this/self",
    [
        lambda core: UIAction(
            ACTION_ID,
            idaapi.action_desc_t(
                ACTION_ID,
                "Update the first function argument to this/self and change its type",
                ThisArgFixerAction(),
                "Ctrl+T",
            ),
            dynamic_menu_add=lambda widget, popup: idaapi.get_widget_type(widget) == idaapi.BWN_DISASM
            or idaapi.get_widget_type(widget) == idaapi.BWN_PSEUDOCODE,
        )
    ],
)


class ThisArgFixerAction(action_handler_t):
    def activate(self, ctx: ida_kernwin.action_ctx_base_t):
        if ctx.cur_func is None:
            print("[Error] Not inside a function")
            return False

        if update_argument(ctx.cur_func) and ctx.widget is not None:
            widgets.refresh_widget(ctx.widget)
        return False

    def update(self, ctx):
        return idaapi.AST_ENABLE_ALWAYS



import ida_hexrays
from ida_hexrays import Hexrays_Hooks, cexpr_t, cfuncptr_t, cinsn_t
from idahelper import memory
from idahelper.ast import cexpr, cinsn

from .block import BlockBaseFieldsAssignments, get_block_type, get_ida_block_lvars
from .block_arg_byref import BlockByRefArgBaseFieldsAssignments, get_block_byref_args_lvars
from .utils import StructFieldAssignment, get_struct_fields_assignments


class objc_blocks_optimizer_hooks_t(Hexrays_Hooks):
    def maturity(self, func: cfuncptr_t, new_maturity: int) -> int:
        if new_maturity < ida_hexrays.CMAT_CPA:
            return 0

        optimize_blocks(func)
        optimize_block_byref_args(func)
        return 0


# region byref args
def optimize_block_byref_args(func: cfuncptr_t) -> bool:
    # Check if the function has blocks
    byref_lvars = get_block_byref_args_lvars(func)
    if not byref_lvars:
        return False

    assignments = get_struct_fields_assignments(func, byref_lvars)
    has_optimized = False
    for lvar, lvar_assignments in assignments.items():
        has_optimized |= optimize_block_byref_arg(lvar, func, lvar_assignments)

    return has_optimized


def optimize_block_byref_arg(lvar: str, func: cfuncptr_t, assignments: list[StructFieldAssignment]) -> bool:
    byref_fields = BlockByRefArgBaseFieldsAssignments.initial()
    for assignment in assignments:
        byref_fields.add_assignment(assignment)

    if not byref_fields.is_completed():
        return False

    new_insn = create_byref_init_insn(lvar, func, byref_fields)
    first_assignment = byref_fields.assignments[0]
    for assignment in byref_fields.assignments[1:]:
        assignment.cleanup()
    first_assignment.swap(new_insn)
    return True


def create_byref_init_insn(lvar: str, func: cfuncptr_t, byref_fields: BlockByRefArgBaseFieldsAssignments) -> cinsn_t:
    if byref_fields.byref_dispose is not None:
        call = cexpr.call_helper_from_sig(
            "_byref_block_arg_ex_init",
            byref_fields.type,
            [
                cexpr_t(byref_fields.flags),
                cexpr_t(byref_fields.byref_keep),
                cexpr_t(byref_fields.byref_dispose),
            ],
        )
    else:
        call = cexpr.call_helper_from_sig(
            "_byref_block_arg_init",
            byref_fields.type,
            [
                cexpr_t(byref_fields.flags),
            ],
        )

    lvar_exp = cexpr.from_var_name(lvar, func)

    return cinsn.from_expr(cexpr.from_assignment(lvar_exp, call), ea=byref_fields.ea)


# endregion


# region blocks
def optimize_blocks(func: cfuncptr_t) -> bool:
    # Check if the function has blocks
    block_lvars = get_ida_block_lvars(func)
    if not block_lvars:
        return False

    assignments = get_struct_fields_assignments(func, block_lvars)
    has_optimized = False
    for lvar, lvar_assignments in assignments.items():
        has_optimized |= optimize_block(lvar, func, lvar_assignments)

    return has_optimized


def optimize_block(lvar: str, func: cfuncptr_t, assignments: list[StructFieldAssignment]) -> bool:
    block_fields = BlockBaseFieldsAssignments.initial()
    for assignment in assignments:
        block_fields.add_assignment(assignment)

    if not block_fields.is_completed():
        return False

    new_insn = create_block_init_insn(lvar, func, block_fields)
    first_assignment = block_fields.assignments[0]
    for assignment in block_fields.assignments[1:]:
        assignment.cleanup()
    first_assignment.swap(new_insn)
    return True


def create_block_init_insn(lvar: str, func: cfuncptr_t, block_fields: BlockBaseFieldsAssignments) -> cinsn_t:
    if (isa := get_isa(block_fields.isa)) is not None:
        call = cexpr.call_helper_from_sig(
            f"_{get_block_type(isa)}_block_init",
            block_fields.type,
            [
                cexpr_t(block_fields.flags),
                cexpr_t(block_fields.descriptor),
                cexpr_t(block_fields.invoke),
            ],
        )
    else:
        call = cexpr.call_helper_from_sig(
            "_block_init",
            block_fields.type,
            [
                cexpr_t(block_fields.isa),
                cexpr_t(block_fields.flags),
                cexpr_t(block_fields.descriptor),
                cexpr_t(block_fields.invoke),
            ],
        )

    lvar_exp = cexpr.from_var_name(lvar, func)

    return cinsn.from_expr(cexpr.from_assignment(lvar_exp, call), ea=block_fields.ea)


def get_isa(isa: cexpr_t) -> str | None:
    """Get the isa name from the isa expression"""
    if isa.op == ida_hexrays.cot_ref:
        inner = isa.x
        if inner.op == ida_hexrays.cot_obj:
            return memory.name_from_ea(inner.obj_ea)
    elif isa.op == ida_hexrays.cot_helper:
        return isa.helper
    return None


# endregion



__all__ = ["objc_sugar_component"]

from ioshelper.base.reloadable_plugin import HexraysHookComponent

from .objc_sugar import objc_selector_hexrays_hooks_t

objc_sugar_component = HexraysHookComponent.factory("ObjcSugar", [objc_selector_hexrays_hooks_t])



__all__ = ["apply_kalloc_type_component", "apply_kalloc_types"]

import ida_kernwin
import idaapi
from ida_kernwin import action_handler_t

from ioshelper.base.reloadable_plugin import UIAction, UIActionsComponent

from .kalloc_type import apply_kalloc_types

ACTION_ID = "ioshelper:apply_kalloc_type"

apply_kalloc_type_component = UIActionsComponent.factory(
    "Locate all the kalloc_type_view in the kernelcache and apply them on types",
    [
        lambda core: UIAction(
            ACTION_ID,
            idaapi.action_desc_t(
                ACTION_ID,
                "Locate all the kalloc_type_view in the kernelcache and apply them on types",
                ApplyKallocTypesAction(),
            ),
            menu_location=UIAction.base_location(core),
        )
    ],
)


class ApplyKallocTypesAction(action_handler_t):
    def activate(self, ctx: ida_kernwin.action_ctx_base_t):
        apply_kalloc_types()
        return 0

    def update(self, ctx):
        return idaapi.AST_ENABLE_ALWAYS



__all__ = ["LogCallInfo", "LogCallParams", "get_call_info_for_name", "log_type_to_str"]

import re
from dataclasses import dataclass

from ioshelper.base.utils import match_dict


@dataclass
class LogCallInfo:
    size_index: int
    buf_index: int
    format_index: int
    type_index: int
    name_index: int
    is_signpost: bool


@dataclass
class LogCallParams:
    log_type: int
    size: int
    stack_base_offset: int
    format_str_ea: int
    call_ea: int
    name_str_ea: int
    is_signpost: bool


OS_LOG_IMPL_CALL_INFO = LogCallInfo(
    type_index=2, format_index=3, buf_index=4, size_index=5, name_index=None, is_signpost=False
)
OS_SIGNPOST_EMIT_WITH_NAME_CALL_INFO = LogCallInfo(
    type_index=2, name_index=4, format_index=5, buf_index=6, size_index=7, is_signpost=True
)

OS_LOG_NAMES: dict[str | re.Pattern, LogCallInfo] = {
    "__os_log_impl": OS_LOG_IMPL_CALL_INFO,
    "__os_log_error_impl": OS_LOG_IMPL_CALL_INFO,
    "__os_log_debug_impl": OS_LOG_IMPL_CALL_INFO,
    "__os_log_info_impl": OS_LOG_IMPL_CALL_INFO,
    re.compile(r"__os_log_impl_(\d+)"): OS_LOG_IMPL_CALL_INFO,
    re.compile(r"__os_log_error_impl_(\d+)"): OS_LOG_IMPL_CALL_INFO,
    re.compile(r"__os_log_debug_impl_(\d+)"): OS_LOG_IMPL_CALL_INFO,
    re.compile(r"__os_log_info_impl(\d+)"): OS_LOG_IMPL_CALL_INFO,
    "__os_signpost_emit_with_name_impl": OS_SIGNPOST_EMIT_WITH_NAME_CALL_INFO,
    re.compile(r"__os_signpost_emit_with_name_impl_(\d+)"): OS_SIGNPOST_EMIT_WITH_NAME_CALL_INFO,
}

LOG_TYPES: dict[int, str] = {0: "default", 1: "info", 2: "debug", 16: "error", 17: "fault"}
SIGNPOST_TYPES: dict[int, str] = {0: "event", 1: "intervalBegin", 2: "intervalEnd"}


def log_type_to_str(log_type: int, is_signpost: bool) -> str:
    if not is_signpost:
        return LOG_TYPES.get(log_type, f"log{log_type}")
    else:
        return SIGNPOST_TYPES.get(log_type, f"signpost{log_type}")


def get_call_info_for_name(name: str) -> LogCallInfo | None:
    return match_dict(OS_LOG_NAMES, name)



__all__ = ["objc_calls_optimizer_t"]

import re

import ida_hexrays
from ida_hexrays import (
    mblock_t,
    mcallinfo_t,
    minsn_t,
    minsn_visitor_t,
    mop_t,
    mop_visitor_t,
)
from idahelper.microcode import minsn, mreg

from ioshelper.base.utils import CounterMixin, match

# Replace f(x) with x
ID_FUNCTIONS_TO_REPLACE_WITH_ARG: list[str | re.Pattern] = [
    "objc_retain",
    "objc_retainAutorelease",
    "objc_autoreleaseReturnValue",
    "objc_autorelease",
    "_objc_claimAutoreleasedReturnValue",
    re.compile(r"_objc_claimAutoreleasedReturnValue_(\d+)"),
    "_objc_retainBlock",
    "objc_unsafeClaimAutoreleasedReturnValue",
    "objc_retainAutoreleasedReturnValue",
]

# Remove f(x) calls
VOID_FUNCTIONS_TO_REMOVE_WITH_SINGLE_ARG: list[str | re.Pattern] = [
    # Objective-C
    "objc_release",
    # intrinsics
    "__break",
    # CFoundation
    "_CFRelease",
    re.compile(r"_CFRelease_(\d+)"),
]

VOID_FUNCTION_TO_REMOVE_WITH_MULTIPLE_ARGS: list[str | re.Pattern] = [
    # Blocks
    "__Block_object_dispose",
    "_Block_object_dispose",
    re.compile(r"__Block_object_dispose_(\d+)"),
]

# Replace assign(&x, y) with x = y;
ASSIGN_FUNCTIONS: list[str | re.Pattern] = [
    "_objc_storeStrong",
    "j__objc_storeStrong",
    re.compile(r"j__objc_storeStrong_(\d+)"),
]


class mop_optimizer_t(mop_visitor_t, CounterMixin):
    def visit_mop(self, op: mop_t, tp, is_target: bool) -> int:
        # No assignment dest, we want a call instruction
        if not is_target and op.d is not None:
            self.visit_instruction_mop(op)
        return 0

    def visit_instruction_mop(self, op: mop_t):
        # We only want calls
        insn: minsn_t = op.d
        if insn.opcode != ida_hexrays.m_call:
            return

        # Calls with names
        name = minsn.get_func_name_of_call(insn)
        if name is None:
            return

        # If it should be optimized to first arg, optimize
        if match(ID_FUNCTIONS_TO_REPLACE_WITH_ARG, name):
            fi: mcallinfo_t = insn.d.f
            if fi.args.empty():
                # No arguments, probably IDA have not optimized it yet
                return

            # Swap mop containing call with arg0
            op.swap(fi.args[0])
            self.count()


class insn_optimizer_t(minsn_visitor_t, CounterMixin):
    def visit_minsn(self) -> int:
        # We only want calls
        insn: minsn_t = self.curins
        if insn.opcode == ida_hexrays.m_call:
            self.visit_call_insn(insn, self.blk)
        return 0

    def visit_call_insn(self, insn: minsn_t, blk: mblock_t):
        # Calls with names
        name = minsn.get_func_name_of_call(insn)
        if name is None:
            return

        for optimization in [
            self.void_function_to_remove,
            self.id_function_to_replace_with_their_arg,
            self.assign_functions,
        ]:
            # noinspection PyArgumentList
            if optimization(name, insn, blk):
                return

    def void_function_to_remove(self, name: str, insn: minsn_t, blk: mblock_t) -> bool:
        if match(VOID_FUNCTIONS_TO_REMOVE_WITH_SINGLE_ARG, name):
            single_arg = True
        elif match(VOID_FUNCTION_TO_REMOVE_WITH_MULTIPLE_ARGS, name):
            single_arg = False
        else:
            return False

        fi: mcallinfo_t = insn.d.f
        if fi.args.empty() or (single_arg and len(fi.args) != 1):
            # No arguments, probably not optimized yet
            # Or not matching the number of arguments
            return False

        if any(arg.has_side_effects() for arg in fi.args):
            print("[Error] arguments with side effects are not supported yet!")
            return False

        if not fi.return_type or not fi.return_type.is_void():
            # embedded instruction, the result can be assigned to something.
            print(
                f"[Error] Cannot remove {name} as this is an embedded instruction. Is the return type correct? it should be void."
            )
            return False

        blk.make_nop(insn)
        self.count()
        return True

    def id_function_to_replace_with_their_arg(self, name: str, insn: minsn_t, _blk: mblock_t) -> bool:
        if not match(ID_FUNCTIONS_TO_REPLACE_WITH_ARG, name):
            return False

        # Might be a call with destination (for example, if it is the last statement in the function)
        fi: mcallinfo_t = insn.d.f
        if fi.args.empty() or fi.retregs.empty():
            # No arguments (probably not optimized yet) or no return reg
            return False

        # Make instruction mov instead of call
        insn.opcode = ida_hexrays.m_mov
        insn.l.swap(fi.args[0])
        insn.d.swap(fi.retregs[0])
        self.count()
        return True

    def assign_functions(self, name: str, insn: minsn_t, _blk: mblock_t) -> bool:
        if not match(ASSIGN_FUNCTIONS, name):
            return False

        fi: mcallinfo_t = insn.d.f
        if fi.args.size() != 2:
            # Not enough argument, probably not optimized yet
            return False
        insn.opcode = ida_hexrays.m_stx
        # src
        insn.l.swap(fi.args[1])
        # dest
        insn.d.swap(fi.args[0])
        # seg - need to be CS/DS according to the docs.
        insn.r.make_reg(mreg.cs_reg(), 2)
        self.count()
        return True


class objc_calls_optimizer_t(ida_hexrays.optinsn_t):
    def func(self, blk: mblock_t, ins: minsn_t, optflags: int):
        # Let IDA reconstruct the calls before
        if blk.mba.maturity < ida_hexrays.MMAT_CALLS:
            return 0

        mop_optimizer = mop_optimizer_t(blk.mba, blk)
        insn_optimizer = insn_optimizer_t(blk.mba, blk)
        ins.for_all_ops(mop_optimizer)
        ins.for_all_insns(insn_optimizer)
        changes = mop_optimizer.cnt + insn_optimizer.cnt
        if changes:
            blk.mark_lists_dirty()
        return changes



__all__ = ["range_condition_optimizer_component"]

from ioshelper.base.reloadable_plugin import HexraysHookComponent

from .range_condition import range_condition_optimizer

range_condition_optimizer_component = HexraysHookComponent.factory(
    "Range condition optimizer", [range_condition_optimizer]
)



__all__ = ["search_field_access"]

from collections.abc import Iterator
from dataclasses import dataclass
from enum import Enum
from typing import Generic, TypeVar

import ida_hexrays
from ida_funcs import func_t
from ida_hexrays import mblock_t, minsn_t, mop_t
from ida_typeinf import tinfo_t
from idahelper import cpp, memory, segments, tif
from idahelper.microcode import mba, mop
from idahelper.microcode.visitors import (
    TreeVisitOrder,
    extended_microcode_visitor_t,
)
from idahelper.segments import Segment

PTR_SIZE = 8


def search_field_access(
    struct_name: str,
    offset: int,
    segments_to_search: list[str],
    show_read: bool,
    show_write: bool,
    filter_types: bool = True,
):
    """
    Search for field access in the given segments.

    `struct_name` - "IOService"
    `offset`: 0x28
    `segments_to_search`: ["com.apple.iokit.IOPortFamily", "com.apple.kernel"]
    `show_read`: Should show read access to the field
    `show_write`: Should show write access to the field
    `filter_types`: Should filter results that we know their type does not match the struct type children.
    """
    typ = tif.from_struct_name(struct_name)
    assert typ is not None
    typ_children: set[str | None] = {f"{t!s} *" for t in tif.get_children_classes(typ) or []}
    typ_children.add(f"{struct_name} *")

    results: list[FieldAccess] = []
    for seg_to_search in segments_to_search:
        seg = _get_text_segment(seg_to_search)
        if seg is None:
            print(f"[Error] segment {seg_to_search} not found")
            continue
        print(f"Searching in {seg.name}")

        for result in _collect_field_accesses(seg):
            # Skip vtable
            if result.prev_access is not None and result.prev_access.offset == 0:
                continue

            if filter_types and result.src_type is not None and str(result.src_type) not in typ_children:
                continue

            if result.offset == offset:
                results.append(result)

        print(f"[Status] Finished scanning in {seg_to_search}!")

    for result in results:
        if (show_read and result.type == AccessType.READ) or (show_write and result.type == AccessType.WRITE):
            # noinspection PyBroadException
            try:
                print(result.compact_str())
            except:  # noqa: E722
                print(f"result: {result.ea}")


def _collect_field_accesses(segment: Segment) -> "Iterator[FieldAccess]":
    for func in segment.functions():
        name = memory.name_from_ea(func.start_ea)

        func_mba = mba.from_func(func.start_ea)
        if func_mba is None:
            print(f"[Error] failed to get mba of func {name}")
            continue

        collector = field_access_collector()
        collector.visit_function(func_mba)
        yield from collector.results


def _get_text_segment(name: str) -> Segment | None:
    """Get the text segment for the given name, or None if not found."""
    seg = segments.get_segment_by_name(f"{name}:__text")
    if seg is None:
        seg = segments.get_segment_by_name(f"{name}:__TEXT_EXEC.__text")
    return seg


class AccessType(Enum):
    READ = 0
    WRITE = 1
    UNKNOWN = 2


@dataclass
class FieldAccess:
    op: mop_t
    offset: int | None
    insn: minsn_t
    top_ins: minsn_t
    ea: int
    blk: mblock_t
    type: AccessType
    prev_access: "FieldAccess | None" = None
    src_type: tinfo_t | None = None

    def __str__(self) -> str:
        try:
            return f"FieldAccess(op={self.op.dstr()}, offset={self.offset}, type: {self.type}, prev_offset: {self.prev_access.offset if self.prev_access is not None else '<none>'}, ea: {self.ea:X}, src_type: {self.src_type}, insn: {self.insn.dstr()}), top_ins: {self.top_ins.dstr()})"
        except:  # noqa: E722
            return self.compact_str()

    def compact_str(self) -> str:
        return f"FieldAccess(offset={self.offset}, type: {self.type}, prev_offset: {self.prev_access.offset if self.prev_access is not None else '<none>'}, ea: {self.ea:X}, src_type: {self.src_type})"


T = TypeVar("T")


class MopDict(Generic[T]):
    def __init__(self):
        self._storage: dict[str, T] = {}

    def __setitem__(self, op: mop_t, obj: T):
        self._storage[op.dstr()] = obj

    def __getitem__(self, op: mop_t):
        return self._storage[op.dstr()]

    def __contains__(self, op: mop_t) -> bool:
        return op.dstr() in self._storage

    def __iter__(self):
        return iter(self._storage.values())

    def __len__(self):
        return len(self._storage)

    def get(self, op: mop_t, default: T) -> T:
        if op in self:
            return self[op]

        return default

    def get_or_none(self, op: mop_t) -> T | None:
        if op in self:
            return self[op]
        return None


# TODO: support globals
# TODO: support write to func result
# icall  cs.2{4}, [cs.2{4}:([cs.2{4}:x2_0.8{5}].8+#0x20.8)].8, <fast:_QWORD x2_0.8{5}>.0
class field_access_collector(extended_microcode_visitor_t):
    def __init__(self):
        super().__init__(TreeVisitOrder.POST_ORDER)
        self.results: list[FieldAccess] = []
        self.mop_to_offset: MopDict[int | None] = MopDict()
        self.mop_to_type: MopDict[tinfo_t | None] = MopDict()
        self.mop_to_prev_access: MopDict[FieldAccess] = MopDict()

    def is_offsetable(self, op: mop_t) -> bool:
        if op in self.mop_to_offset:
            return True

        return mop.get_local_variable(op) is not None

    def get_offset(self, op: mop_t) -> int | None:
        val = self.mop_to_offset.get_or_none(op)
        if val is not None:
            return val

        if mop.get_local_variable(op) is not None:
            return 0

        return None

    def get_prev_field_access(self, op: mop_t) -> FieldAccess | None:
        return self.mop_to_prev_access.get_or_none(op)

    def get_type(self, op: mop_t) -> tinfo_t | None:
        return self.mop_to_type.get_or_none(op)

    def track_offsetable_parent_mop(self, offset: int | None):
        """We were in instruction, which created a possible pointer. We want to mark the embedding mop as possible pointer"""
        if self.parents:
            assert isinstance(self.parents[-1], mop_t)
            self.mop_to_offset[self.parents[-1]] = offset

    def track_parent_type(self, typ: tinfo_t | None):
        """We were in instruction, which created a possible pointer. We want to mark the embedding mop as possible pointer"""
        if self.parents:
            assert isinstance(self.parents[-1], mop_t)
            self.mop_to_type[self.parents[-1]] = typ

    def track_prev_field_access(self, field_access: FieldAccess | None):
        if self.parents and field_access is not None:
            assert isinstance(self.parents[-1], mop_t)
            self.mop_to_prev_access[self.parents[-1]] = field_access

    def _visit_insn(self, ins: minsn_t) -> int:
        # print('  ' * len(self.parents), "MOP:", op.dstr())
        # print("  " * len(self.parents), "INS:", ins.dstr())

        if ins.opcode == ida_hexrays.m_add:
            return self._visit_add(ins)
        elif ins.opcode == ida_hexrays.m_ldx:
            return self._visit_ldx(ins)
        elif ins.opcode == ida_hexrays.m_stx:
            return self._visit_stx(ins)
        return 0

    def _visit_ldx(self, ins: minsn_t) -> int:
        if self.is_offsetable(ins.r):
            offset = self.get_offset(ins.r)

            access = FieldAccess(
                ins.r,
                offset,
                ins,
                self.top_ins,
                ins.ea,
                self.blk,
                AccessType.READ,
                self.get_prev_field_access(ins.r),
                self.get_type(ins.r),
            )
            self.results.append(access)
            self.track_offsetable_parent_mop(0)
            self.track_prev_field_access(access)

        return 0

    def _visit_stx(self, ins: minsn_t) -> int:
        if self.is_offsetable(ins.d):
            offset = self.get_offset(ins.d)

            access = FieldAccess(
                ins.d,
                offset,
                ins,
                self.top_ins,
                ins.ea,
                self.blk,
                AccessType.WRITE,
                self.get_prev_field_access(ins.d),
                self.get_type(ins.d),
            )
            self.results.append(access)
            self.track_offsetable_parent_mop(0)
            self.track_prev_field_access(access)
        return 0

    def _visit_add(self, ins: minsn_t) -> int:  # noqa: C901
        """Search for `add local, offset, .8`"""

        # Check output is of pointer size
        if ins.d.size != PTR_SIZE:
            return 0

        # Check if one of them is a possible pointer
        if self.is_offsetable(ins.l):
            ptr = ins.l
            offset = mop.get_const_int(ins.r)
        elif self.is_offsetable(ins.r):
            ptr = ins.r
            offset = mop.get_const_int(ins.l)
        else:
            return 0

        # Check if the offset is a const
        if offset is None:
            # print("Dynamic access!", hex(ins.ea), ins.dstr())
            return 0

        if not self.parents:
            # The add instruction is direct instruction in the block
            # print(
            #     f"TODO: {hex(ins.ea)} {ins.dstr()} is direct instruction in the block, need to follow destination to understand if it is a load or store"
            # )

            # TODO: hack
            self.mop_to_offset[ins.d] = offset
            return 0

        # Since this is an embedded instruction, there are at least 2 parents.
        # We can skip the first parent as it is an "instruction result" mop.
        parent = self.parents[-1]
        parent_of_parent = self.parents[-2]
        if isinstance(parent_of_parent, mop_t):
            if parent_of_parent.t != ida_hexrays.mop_f:
                print(f"TODO: parent of parent is not instruction {hex(self.top_ins.ea)}: {self.top_ins.dstr()}")
                print("pp:", parent_of_parent.t, parent_of_parent.dstr())
                return 0
            else:
                # Parameter of function, can be ignored
                return 0

        # TODO: get rid of this if else, it is here as sanity
        self.track_offsetable_parent_mop(offset)
        self.track_prev_field_access(self.get_prev_field_access(ptr))

        lvar = mop.get_local_variable(ptr)
        if lvar is not None and lvar.is_arg_var:
            if str(lvar.type()) not in ["_QWORD", "__int64"]:
                self.track_parent_type(lvar.type())  # type: ignore  # noqa: PGH003
            elif lvar == self.mba.arg(0):
                f: func_t = self.mba.get_curfunc()
                f_name = memory.name_from_ea(f.start_ea)
                if f_name is not None:
                    cls_name = cpp.demangle_class_only(f_name)
                    if cls_name is not None:
                        cls_typ = tif.from_struct_name(cls_name)
                        if cls_typ is not None:
                            self.track_parent_type(cls_typ)

        if parent_of_parent.opcode == ida_hexrays.m_stx:
            if parent != parent_of_parent.d:
                # We would not be the selector (because it will be so weird...), so we must be what we load.
                # print(
                #     f"TODO: {self.top_ins.dstr()} stores addition result to variable. Can we rule out this is normal addition and not pointer logic?"
                # )
                pass
        elif parent_of_parent.opcode == ida_hexrays.m_ldx:  # noqa: SIM102
            if parent != parent_of_parent.r:
                # We would not be the selector (because it will be so weird...), and I don't think we can ever be the destination as it have to be a register.
                print(f"TODO: {self.top_ins.dstr()} load to complex expression. How can it be?")

        return 0



import time
from functools import cache

import ida_hexrays
from ida_hexrays import Hexrays_Hooks, cfunc_t
from idahelper import xrefs
from idahelper.microcode import mba

from .handlers import GLOBAL_HANDLERS, LOCAL_HANDLERS
from .renamer import (
    Modifications,
)
from .visitor import FuncXref, XrefsMatcher, process_function_calls

ALL_HANDLERS = [*LOCAL_HANDLERS, *GLOBAL_HANDLERS]


@cache
def get_global_xref_matcher() -> XrefsMatcher:
    """Get a matcher for global xrefs."""
    callbacks = []
    for handler in GLOBAL_HANDLERS:
        source_xref = handler.get_source_xref()
        if source_xref is None or not isinstance(source_xref, FuncXref):
            continue
        callbacks.append((source_xref, handler.on_call))
    # noinspection PyTypeChecker
    return XrefsMatcher.build(callbacks)


@cache
def get_all_xref_matcher() -> XrefsMatcher:
    """Get a matcher for global xrefs."""
    callbacks = []
    for handler in ALL_HANDLERS:
        source_xref = handler.get_source_xref()
        if source_xref is None:
            continue
        callbacks.append((source_xref, handler.on_call))
    # noinspection PyTypeChecker
    return XrefsMatcher.build(callbacks)


def apply_global_rename():
    before = time.time()
    for i, handler in enumerate(GLOBAL_HANDLERS):
        print(f"Applying global rename {i + 1}/{len(GLOBAL_HANDLERS)}: {handler.name}")
        source_xref = handler.get_source_xref()
        if source_xref is None or not isinstance(source_xref, FuncXref):
            print(f"Function {handler.name} has no global source xref, skipping")
            continue
        func_ea = source_xref.ea

        xrefs_in_funcs = xrefs.func_xrefs_to(func_ea)
        if not xrefs_in_funcs:
            print(f"Function {handler.name} not called")
            continue

        print(f"Found {len(xrefs_in_funcs)} functions that call {handler.name}:")
        for j, xref_func_ea in enumerate(xrefs_in_funcs):
            with Modifications(xref_func_ea, func_lvars=None) as modifications:
                print(f"  {j + 1}/{len(xrefs_in_funcs)}: {xref_func_ea:#x}")
                process_function_calls(mba.from_func(xref_func_ea), get_global_xref_matcher(), modifications)
    after = time.time()
    print(f"Completed! Took {int(after - before)} seconds")


class LocalRenameHooks(Hexrays_Hooks):
    def maturity(self, cfunc: cfunc_t, new_maturity: int) -> int:
        # For some reason this maturity level is required for typing to be applied for local variables.
        if new_maturity != ida_hexrays.CMAT_CASTED:
            return 0

        with Modifications(cfunc.entry_ea, func_lvars=cfunc.get_lvars()) as modifications:
            process_function_calls(cfunc.mba, get_all_xref_matcher(), modifications)

        return 0


def hooks():
    # Load cache
    get_all_xref_matcher()
    return LocalRenameHooks()



__all__ = ["component"]

from ioshelper.base.reloadable_plugin import OptimizersComponent

from .optimizer import objc_calls_optimizer_t as optimizer

component = OptimizersComponent.factory("Obj-C refcount optimizer", [optimizer])



"""
This is a stub file to be dropped in IDA plugins directory (usually ~/.idapro/plugins)
You should install ida-ios-helper package globally in your python installation (When developing, use an editable install...)
Make sure that this is the python version that IDA is using (otherwise you can switch with idapyswitch...)
Then copy:
- ida_plugin_stub.py to ~/idapro/plugins/ida_ios_helper/ida_plugin_stub.py
- ida-plugin.json to ~/idapro/plugins/ida_ios_helper/ida_plugin.json
"""

# noinspection PyUnresolvedReferences
__all__ = ["PLUGIN_ENTRY", "iOSHelperPlugin"]
try:
    from ioshelper.ida_plugin import PLUGIN_ENTRY, iOSHelperPlugin
except ImportError:
    print("[Error] Could not load ida-ios-helper plugin. ida-ios-helper Python package doesn't seem to be installed.")



from idahelper import functions, memory, strings, widgets, xrefs


def find_matching_string(target_str: str) -> list[tuple[int, str]]:
    matches: list[tuple[int, str]] = []
    for s in strings.strings():
        if target_str in str(s):
            matches.append((s.ea, str(s)))  # type: ignore  # noqa: PGH003
    return matches


def show_xrefs_to_string(ea: int):
    s_xrefs = xrefs.get_xrefs_to(ea)
    if not s_xrefs:
        print("No cross-references found.")
    elif len(s_xrefs) == 1:
        widgets.jump_to(s_xrefs[0])
    else:
        print("Multiple xrefs to the string:")
        for xref in s_xrefs:
            print_xref(xref)


def print_xref(ea: int, match: str | None = None):
    func_start = functions.get_start_of_function(ea)
    if func_start is None:
        print(f"{ea:X}")
        return

    func_name = memory.name_from_ea(func_start) or "<unknown>"
    match_str = "" if match is None else f": {match}"

    print(f"{ea:X} at {func_name}+{ea - func_start:X}{match_str}")


def jump_to_string_ask():
    target_str = widgets.show_string_input("Enter substring to search in binary strings")
    if not target_str:
        return

    matches = find_matching_string(target_str)

    if not matches:
        print("[Warning] No matching strings found.")
        return

    # If there's only one result, or an exact match, show xrefs
    if len(matches) == 1 or any(s == target_str for _, s in matches):
        for ea, s in matches:
            if s == target_str or len(matches) == 1:
                show_xrefs_to_string(ea)
                return

    # Otherwise, let the user choose
    print("Multiple results for the string:")
    for ea, s in matches:
        print(f"{ea:X}: {s}")



__all__ = ["GLOBAL_HANDLERS", "LOCAL_HANDLERS"]


from idahelper import file_format, functions, instructions, memory, tif, xrefs

from .renamer import (
    FuncHandler,
    FuncHandlerByNameWithStringFinder,
    FuncHandlerVirtualGetter,
    FuncHandlerVirtualSetter,
    Modifications,
)
from .visitor import Call, FuncXref, SourceXref


class OSSymbolHandler(FuncHandlerByNameWithStringFinder):
    def __init__(self, name: str, search_string: str):
        super().__init__(
            name,
            tif.from_func_components("OSSymbol*", [tif.FuncParam("const char*", "cString")]),
            search_string,
            is_call=True,
        )

        self._cached_symbol_type = tif.from_c_type("OSSymbol*")

    def on_call(self, call: Call, modifications: Modifications):
        self._rename_assignee_by_index(modifications, call, 0, modifier=lambda name: f"sym_{name}")
        self._retype_assignee(modifications, call, self._cached_symbol_type)


OSSymbol_WithCStringNoCopy = OSSymbolHandler("OSSymbol::withCStringNoCopy", "IOMatchedPersonality")
OSSymbol_WithCString = OSSymbolHandler("OSSymbol::withCString", "ACIPCInterfaceProtocol")


class IORegistry_MakePlane(FuncHandlerByNameWithStringFinder):
    def __init__(self):
        super().__init__(
            "IORegistry::makePlane",
            tif.from_func_components("IORegistryPlane*", [tif.FuncParam("const char*", "name")]),
            "ChildLinks",
            is_call=False,
        )
        self._cached_registry_plane_type = tif.from_c_type("IORegistryPlane*")

    def on_call(self, call: Call, modifications: Modifications):
        self._rename_assignee_by_index(modifications, call, 0, modifier=lambda name: f"plane_{name}")
        self._retype_assignee(modifications, call, self._cached_registry_plane_type)


IOService_SetProperty = [
    FuncHandlerVirtualSetter(
        "IORegistryEntry::setProperty",
        tif.from_c_type("IORegistryEntry"),
        offset,
        name_index=1,
        rename_index=2,
        rename_prefix="val_",
    )
    for offset in range(0xB8, 0xF0, 8)
]
IOService_GetProperty = FuncHandlerVirtualGetter(
    "IORegistryEntry::getProperty", tif.from_c_type("IORegistryEntry"), 0x118, name_index=1, rename_prefix="val_"
)
IOService_CopyProperty = FuncHandlerVirtualGetter(
    "IORegistryEntry::copyProperty", tif.from_c_type("IORegistryEntry"), 0x148, name_index=1, rename_prefix="val_"
)


class MetaClassConstructor(FuncHandlerByNameWithStringFinder):
    def __init__(self):
        super().__init__(
            "__ZN11OSMetaClassC2EPKcPKS_j",
            tif.from_func_components(
                "OSMetaClass*",
                [
                    tif.FuncParam("OSMetaClass*", "this"),
                    tif.FuncParam("const char*", "className"),
                    tif.FuncParam("const OSMetaClass*", "superClass"),
                    tif.FuncParam("unsigned int", "classSize"),
                ],
            ),
            "OSMetaClass: preModLoad() wasn't called for class %s (runtime internal error).",
            is_call=False,
        )

        self._cached_metaclass_type = tif.from_c_type("OSMetaClass*")

    def on_call(self, call: Call, modifications: Modifications):
        self._rename_parameter_by_index(
            modifications,
            call,
            name_index=1,
            rename_index=0,
            modifier=lambda name: f"__ZN{len(name)}{name}10gMetaclassE",
        )
        self._retype_parameter_by_index(modifications, call, 0, self._cached_metaclass_type)


class peParseBootArgn(FuncHandler):
    def __init__(self):
        super().__init__("PE_parse_boot_argn")

    def get_source_xref(self) -> SourceXref | None:
        existing = memory.ea_from_name("PE_parse_boot_argn")
        if existing is not None:
            return FuncXref(existing)

    def on_call(self, call: Call, modifications: Modifications):
        self._rename_parameter_by_index(
            modifications, call, 0, 1, modifier=lambda name: f"boot_{name.replace('-', '_')}"
        )

        if isinstance(call.params[2], int):
            new_type = tif.from_size(call.params[2])
            if new_type is not None:
                self._retype_parameter_by_index(modifications, call, 1, tif.pointer_of(new_type))


class StackCheckFail(FuncHandler):
    def __init__(self):
        # For some reason I cannot set the name of the function to the original name, as IDA hides call to the function
        # So we use a different name
        super().__init__("__xnu_stack_check_fail")

    def get_source_xref(self) -> SourceXref | None:
        existing = memory.ea_from_name(self.name)
        if existing is not None:
            return FuncXref(existing)
        searched = list(xrefs.string_xrefs_to("stack_protector.c"))
        if not searched:
            print("[Error] Could not find xrefs to 'stack_protector.c' for", self.name)
            return None
        ldr_addr = searched[0]

        func_start_ea = StackCheckFail.get_previous_pacibsp(ldr_addr)
        func_end_ea = StackCheckFail.get_after_next_bl(ldr_addr)
        if func_start_ea is None or func_end_ea is None:
            print("[Error] Could not find function boundaries:", self.name)
            return None

        if not functions.is_in_function(func_start_ea) and not functions.add_function(func_start_ea, func_end_ea):
            print(f"[Error] Could not add function {self.name} at {func_start_ea:#x}")
            return None

        if not tif.apply_tinfo_to_ea(tif.from_func_components("void", [tif.FuncParam("void")]), func_start_ea):
            print(f"[Error] Could not apply tinfo to function {self.name} at {func_start_ea:#x}")
            return None

        if not functions.apply_flag_to_function(func_start_ea, functions.FLAG_NO_RETURN):
            print(f"[Error] Could not apply no-return flag to function {self.name} at {func_start_ea:#x}")
            return None

        if not memory.set_name(func_start_ea, self.name, retry=True):
            print(f"[Error] Could not set name for function {self.name} at {func_start_ea:#x}")
            return None

        return FuncXref(func_start_ea)

    @staticmethod
    def get_previous_pacibsp(call_ea: int) -> int | None:
        """Given a call, search previous instructions to find a movk call"""
        insn = instructions.decode_instruction(call_ea)
        if not insn:
            return None

        for _ in range(10):
            insn = instructions.decode_previous_instruction(insn)
            # No more instructions in this execution flow
            if insn is None:
                break
            if insn.get_canon_mnem() == "PAC":
                return insn.ea
        return None

    @staticmethod
    def get_after_next_bl(call_ea: int) -> int | None:
        """Given a call, search previous instructions to find a movk call"""
        insn = instructions.decode_instruction(call_ea)
        if not insn:
            return None

        for _ in range(10):
            insn = instructions.decode_instruction(insn.ea + insn.size)
            # No more instructions in this execution flow
            if insn is None:
                break
            if insn.get_canon_mnem() == "BL":
                return insn.ea + insn.size
        return None

    def on_call(self, call: Call, modifications: Modifications):
        # Do nothing on call, we just want to rename the function
        pass


if file_format.is_kernelcache():
    GLOBAL_HANDLERS: list[FuncHandler] = [
        OSSymbol_WithCStringNoCopy,
        OSSymbol_WithCString,
        IORegistry_MakePlane(),
        MetaClassConstructor(),
    ]
    LOCAL_HANDLERS: list[FuncHandler] = [
        *IOService_SetProperty,
        IOService_GetProperty,
        IOService_CopyProperty,
        peParseBootArgn(),
        StackCheckFail(),
    ]
else:
    GLOBAL_HANDLERS = []
    LOCAL_HANDLERS = []



from typing import Literal

import ida_hexrays
from ida_hexrays import Hexrays_Hooks, cexpr_t, cfunc_t
from idahelper import tif
from idahelper.ast import cexpr


def is_unsigned_comparison_expr(e: cexpr_t) -> bool:
    """Check if the expression is an unsigned comparison."""
    return e.op in (ida_hexrays.cot_uge, ida_hexrays.cot_ule, ida_hexrays.cot_ugt, ida_hexrays.cot_ult)


class RangeConditionTreeVisitor(ida_hexrays.ctree_visitor_t):
    def __init__(self, func: cfunc_t):
        super().__init__(ida_hexrays.CV_FAST)
        self.func = func

    def visit_expr(self, e: cexpr_t) -> int:
        # search for pattern of x ± const_1 < const_2
        if not is_unsigned_comparison_expr(e):
            return 0

        lhs, rhs = e.x, e.y
        if e.x.op == ida_hexrays.cot_cast:
            is_cast = True
            lhs = e.x.x
        else:
            is_cast = False

        if (
            rhs.op != ida_hexrays.cot_num
            or lhs.op not in (ida_hexrays.cot_add, ida_hexrays.cot_sub)
            or lhs.y.op != ida_hexrays.cot_num
            or lhs.x.has_side_effects()
        ):
            return 0

        # Get modulus for the expression size - use e.x instead of lhs to handle casts correctly
        expr_size_in_bytes = e.x.type.get_size()
        if expr_size_in_bytes not in (1, 2, 4, 8):
            print(f"[Warning] Unsupported expression size {expr_size_in_bytes} for {e.dstr()}")
            return 0
        mod = 1 << (expr_size_in_bytes << 3)

        # Get consts
        lhs_const = lhs.y.numval() % mod
        rhs_const = e.y.numval() % mod

        # Complement lhs const if the operation is addition
        if lhs.op == ida_hexrays.cot_add:
            lhs_const = mod - lhs_const

        x = cexpr.from_cast(lhs.x, e.x.type) if is_cast else lhs.x

        replacement_expr = (
            create_range_condition_greater_than
            if e.op in (ida_hexrays.cot_ugt, ida_hexrays.cot_uge)
            else create_range_condition_less_than
        )(e, lhs_const, rhs_const, mod, x, lhs.y.ea, self.func)
        e.swap(replacement_expr)

        self.prune_now()
        return 0


def create_range_condition_less_than(
    e: cexpr_t, lhs: int, rhs: int, mod: int, x: cexpr_t, lhs_ea: int, func: cfunc_t
) -> cexpr_t:
    """Create a range condition for the expression `x - lhs < rhs`."""

    lhs_plus_rhs = lhs + rhs
    lhs_plus_rhs_mod_n = lhs_plus_rhs % mod
    # if lhs + rhs < mod, we can use a single range condition
    #   x - lhs < rhs  => x ∈ [lhs, lhs + rhs) ==> lhs < x && x < lhs + rhs
    #   x - lhs <= rhs => x ∈ [lhs, lhs + rhs] ==> lhs < x && x <= lhs + rhs
    # if lhs + rhs >= mod, we need to use two range conditions
    #   x - lhs < rhs  => x ∈ [lhs, mod) U x ∈ [0, (lhs+rhs) mod n) ==> lhs < x || x < (lhs + rhs) mod n
    #   x - lhs <= rhs => x ∈ [lhs, mod) U x ∈ [0, (lhs+rhs) mod n] ==> lhs < x || x <= (lhs + rhs) mod n
    # Notice that the conditions are the same, but it is either && or || depending on whether lhs + rhs < mod or not.

    op: Literal["&&", "||"] = "||" if lhs_plus_rhs >= mod else "&&"
    lhs_plus_rhs_expr = cexpr.from_const_value(lhs_plus_rhs_mod_n, func, lhs_ea)
    lhs_expr = cexpr.from_const_value(lhs, func, e.y.ea)
    return _bin_op(
        _bin_op(lhs_expr, "<", cexpr_t(x), e.ea),
        op,
        _bin_op(cexpr_t(x), IDA_OP_TO_MATH_OP[e.op], lhs_plus_rhs_expr, e.ea),
        e.ea,
    )


def create_range_condition_greater_than(
    e: cexpr_t, lhs: int, rhs: int, mod: int, x: cexpr_t, lhs_ea: int, func: cfunc_t
) -> cexpr_t:
    """Create a range condition for the expression `x - lhs > rhs`."""

    lhs_plus_rhs = lhs + rhs
    lhs_plus_rhs_mod_n = lhs_plus_rhs % mod

    lhs_plus_rhs_expr = cexpr.from_const_value(lhs_plus_rhs_mod_n, func, lhs_ea)
    lhs_expr = cexpr.from_const_value(lhs, func, e.y.ea)
    op: Literal["<", "<="] = "<" if e.op == ida_hexrays.cot_ugt else "<="

    # if lhs + rhs < mod:
    #   x - lhs > rhs  => x ∈ (lhs + rhs, mod) U x ∈ [0, lhs) ==> lhs + rhs < x || x < lhs
    #   x - lhs >= rhs => x ∈ [lhs + rhs, mod) U x ∈ [0, lhs] ==> lhs + rhs <= x || x < lhs
    if lhs_plus_rhs < mod:
        return _bin_op(
            _bin_op(lhs_plus_rhs_expr, op, cexpr_t(x), e.ea),
            "||",
            _bin_op(cexpr_t(x), "<", lhs_expr, e.ea),
            e.ea,
        )
    else:
        # if lhs + rhs >= mod:
        #   x - lhs > rhs  => x ∈ (lhs + rhs (mod n), lhs) ==> lhs + rhs (mod n) < x && x < lhs
        #   x - lhs >= rhs => x ∈ [lhs + rhs (mod n), lhs) ==> lhs + rhs (mod n) <= x && x < lhs
        return _bin_op(
            _bin_op(lhs_plus_rhs_expr, op, cexpr_t(x), e.ea),
            "&&",
            _bin_op(cexpr_t(x), "<", lhs_expr, e.ea),
            e.ea,
        )


def _bin_op(left: cexpr_t, op: Literal["<", "<=", ">", ">=", "&&", "||"], right: cexpr_t, ea: int) -> cexpr_t:
    """Create a boolean binary operation expression."""
    if op == "&&":
        ida_op = ida_hexrays.cot_land
    elif op == "||":
        ida_op = ida_hexrays.cot_lor
    else:
        ida_op = MATH_OP_TO_IDA_OP[op]
    return cexpr.from_binary_op(left, right, ida_op, tif.BOOL, ea)


IDA_OP_TO_MATH_OP: dict[int, Literal["<", "<=", ">", ">="]] = {
    ida_hexrays.cot_uge: ">=",
    ida_hexrays.cot_ule: "<=",
    ida_hexrays.cot_ugt: ">",
    ida_hexrays.cot_ult: "<",
}
MATH_OP_TO_IDA_OP = {v: k for k, v in IDA_OP_TO_MATH_OP.items()}


class range_condition_optimizer(Hexrays_Hooks):
    def maturity(self, cfunc: cfunc_t, new_maturity: int) -> int:
        if new_maturity != ida_hexrays.CMAT_NICE:
            return 0

        RangeConditionTreeVisitor(cfunc).apply_to(cfunc.body, None)  # pyright: ignore [reportArgumentType]
        return 0



__all__ = ["apply_kalloc_types"]

from ida_typeinf import tinfo_t
from idahelper import memory, segments, tif

KALLOC_TYPE_DEFINITIONS = """
struct zone_view {
    void*          zv_zone;
    void*    zv_stats;
    const char     *zv_name;
    void*     zv_next;
};

enum kalloc_type_flags_t : uint32_t {
    KT_DEFAULT        = 0x0001,
    KT_PRIV_ACCT      = 0x0002,
    KT_SHARED_ACCT    = 0x0004,
    KT_DATA_ONLY      = 0x0008,
    KT_VM             = 0x0010,
    KT_CHANGED        = 0x0020,
    KT_CHANGED2       = 0x0040,
    KT_PTR_ARRAY      = 0x0080,
    KT_NOSHARED       = 0x2000,
    KT_SLID           = 0x4000,
    KT_PROCESSED      = 0x8000,
    KT_HASH           = 0xffff0000,
};

struct kalloc_type_view {
    struct zone_view        kt_zv;
    const char             *kt_signature;
    kalloc_type_flags_t     kt_flags;
    uint32_t                kt_size;
    void                   *unused1;
    void                   *unused2;
};

struct kalloc_type_var_view {
	uint16_t                kt_version;
	uint16_t                kt_size_hdr;
	uint32_t                kt_size_type;
	void                   *kt_stats;
	const char             *kt_name;
	void                   *kt_next;
	uint16_t                kt_heap_start;
	uint8_t                 kt_zones[22];
	const char             *kt_sig_hdr;
	const char             *kt_sig_type;
	kalloc_type_flags_t     kt_flags;
}
"""
KALLOC_TYPE_VIEW_OFFSET_NAME = 16  # void *zv_zone + void *zv_stats
KALLOC_TYPE_VIEW_OFFSET_SIGNATURE = 32  # zone_view

KALLOC_TYPE_VAR_VIEW_OFFSET_NAME = (
    16  # uint16_t kt_version + uint16_t kt_size_hdr + uint32_t kt_size_type + void* kt_stats
)

VOID_PTR_TYPE: tinfo_t = tif.from_c_type("void*")  # type: ignore   # noqa: PGH003


def apply_kalloc_types():
    kalloc_type_view_tif = tif.from_struct_name("kalloc_type_view")
    kalloc_type_var_view_tif = tif.from_struct_name("kalloc_type_var_view")
    if kalloc_type_view_tif is None or kalloc_type_var_view_tif is None:
        if not tif.create_from_c_decl(KALLOC_TYPE_DEFINITIONS):
            print("[Error] failed to created kalloc_type_view type")

        kalloc_type_view_tif = tif.from_struct_name("kalloc_type_view")
        kalloc_type_var_view_tif = tif.from_struct_name("kalloc_type_var_view")
        if kalloc_type_view_tif is None or kalloc_type_var_view_tif is None:
            print("[Error] could not find kalloc type view")
            return

    classes_handled: set[str] = set()
    for segment in segments.get_segments():
        if segment.name.endswith("__kalloc_type"):
            set_kalloc_type_for_segment(segment, kalloc_type_view_tif, classes_handled)
        if segment.name.endswith("__kalloc_var"):
            set_kalloc_var_for_segment(segment, kalloc_type_var_view_tif)


def set_kalloc_type_for_segment(segment: segments.Segment, kalloc_type_view_tif: tinfo_t, classes_handled: set[str]):
    kalloc_type_view_size = kalloc_type_view_tif.get_size()

    if segment.size % kalloc_type_view_size != 0:
        print(
            f"[Warning] {segment.name} at {segment.start_ea:X} is not a multiple of kalloc_type_view size. is: {segment.size}, not multiple of: {kalloc_type_view_size}"
        )
        return

    for kty_ea in range(segment.start_ea, segment.end_ea, kalloc_type_view_size):
        if not tif.apply_tinfo_to_ea(kalloc_type_view_tif, kty_ea):
            print(f"[Error] failed to apply kalloc_type_view on {kty_ea:X}")

        site_name_ea = memory.qword_from_ea(kty_ea + KALLOC_TYPE_VIEW_OFFSET_NAME)
        site_name = memory.str_from_ea(site_name_ea)
        if site_name is None:
            print(f"[Error] failed to read name for kalloc_type_view on {kty_ea:X}")
            continue

        if not site_name.startswith("site."):
            print(f"[Error] invalid site name on {kty_ea:X}, is: {site_name!r}")
            continue

        class_name = site_name[5:]
        if class_name.startswith("struct "):
            class_name = class_name[7:]
        if class_name.startswith("typeof(") or class_name == "T":
            # Clang generates them using macro, so it might lead to some eccentric specific ones...
            continue

        new_name = f"{escape_name(class_name)}_kty"
        if not memory.set_name(kty_ea, new_name, retry=True, retry_count=50):
            print(f"[Error] failed to rename kalloc_type_view on {kty_ea:X} to {new_name!r}")
            continue

        signature_ea = memory.qword_from_ea(kty_ea + KALLOC_TYPE_VIEW_OFFSET_SIGNATURE)
        signature = memory.str_from_ea(signature_ea)
        if signature is None:
            print(f"[Error] failed to read signature for {new_name} on {kty_ea:X}")
            continue

        try_enrich_type(class_name, signature, classes_handled)


def try_enrich_type(class_name: str, signature: str, classes_handled: set[str]):
    # Don't try to enrich the same type multiple times
    if class_name in classes_handled:
        return
    classes_handled.add(class_name)

    class_tif = tif.from_struct_name(class_name)
    if class_tif is None:
        return

    class_base_offset = tif.get_base_offset_for_class(class_tif)
    if class_base_offset is None:
        return

    # Align it to multiplication of 8
    class_base_offset = int((class_base_offset + 7) / 8) * 8

    # The signature is on 8 bytes each time
    for i in range(class_base_offset // 8, len(signature)):
        c = signature[i]
        if c != "1":
            continue

        member = tif.get_member(class_tif, i * 8)
        if member is None:
            print(f"[Error] {class_name} has no member for at offset {i * 8:#X}")
            return

        # I check for if this is a pointer, because if it was union of data + pointer, the compiler should have yelled
        # So if you changed the type to non pointer, you were mistaken...
        if not member.type.is_ptr() and not tif.set_udm_type(class_tif, member, VOID_PTR_TYPE):
            print(f"[Error] failed to set type for {class_name} member at offset {member.offset}")


def set_kalloc_var_for_segment(segment: segments.Segment, kalloc_type_var_view_tif: tinfo_t):
    kalloc_type_view_size = kalloc_type_var_view_tif.get_size()

    if segment.size % kalloc_type_view_size != 0:
        print(
            f"[Warning] {segment.name} at {segment.start_ea:X} is not a multiple of kalloc_type_view size. is: {segment.size}, not multiple of: {kalloc_type_view_size}"
        )
        return

    for kty_ea in range(segment.start_ea, segment.end_ea, kalloc_type_view_size):
        if not tif.apply_tinfo_to_ea(kalloc_type_var_view_tif, kty_ea):
            print(f"[Error] failed to apply kalloc_type_view on {kty_ea:X}")

        site_name_ea = memory.qword_from_ea(kty_ea + KALLOC_TYPE_VAR_VIEW_OFFSET_NAME)
        site_name = memory.str_from_ea(site_name_ea)
        if site_name is None:
            print(f"[Error] failed to read name for kalloc_type_view on {kty_ea:X}")
            continue

        if not site_name.startswith("site."):
            print(f"[Error] invalid site name on {kty_ea:X}, is: {site_name!r}")
            continue

        class_name = site_name[5:]
        if class_name.startswith("struct "):
            class_name = class_name[7:]
        if class_name.startswith("typeof(") or class_name == "T":
            # Clang generates them using macro, so it might lead to some eccentric specific ones...
            continue

        new_name = f"{escape_name(class_name)}_kty"
        if not memory.set_name(kty_ea, new_name, retry=True, retry_count=50):
            print(f"[Error] failed to rename kalloc_type_view on {kty_ea:X} to {new_name!r}")
            continue


def escape_name(site_name: str) -> str:
    return site_name.replace("*", "_ptr_").replace(" ", "_").replace("<", "_").replace(">", "_")



__all__ = [
    "clang_block_args_analyzer_component",
    "clang_block_optimizer_component",
    "run_objc_plugin_on_func",
    "try_add_block_arg_byref_to_func",
]

import ida_kernwin
import idaapi
from idahelper import widgets

from ioshelper.base.reloadable_plugin import HexraysHookComponent, UIAction, UIActionsComponent

from .analyze_byref_args import try_add_block_arg_byref_to_func
from .optimize_blocks_init import objc_blocks_optimizer_hooks_t
from .utils import run_objc_plugin_on_func

ACTION_ID = "ioshelper:restore_llvm_block_args_byref"

clang_block_args_analyzer_component = UIActionsComponent.factory(
    "Clang Blocks - __block arguments",
    [
        lambda core: UIAction(
            ACTION_ID,
            idaapi.action_desc_t(
                ACTION_ID,
                "Analyze stack-allocated blocks and their __block args (current function)",
                ClangBlockDetectByrefAction(),
                "Alt+Shift+s",
            ),
            menu_location=UIAction.base_location(core),
        )
    ],
)

clang_block_optimizer_component = HexraysHookComponent.factory(
    "Clang Blocks - optimizer", [objc_blocks_optimizer_hooks_t]
)


class ClangBlockDetectByrefAction(ida_kernwin.action_handler_t):
    def activate(self, ctx: ida_kernwin.action_ctx_base_t):
        if ctx.cur_func is None:
            print("No function selected")
            return 0

        run_objc_plugin_on_func(ctx.cur_ea)
        widgets.refresh_pseudocode_widgets()
        if try_add_block_arg_byref_to_func(ctx.cur_func.start_ea):
            widgets.refresh_pseudocode_widgets()
        return 0

    def update(self, ctx) -> int:
        return idaapi.AST_ENABLE_ALWAYS



__all__ = ["update_argument"]

from ida_funcs import func_t
from idahelper import cpp, memory, objc, tif


def update_argument(func: func_t) -> bool:
    func_name = memory.name_from_ea(func.start_ea)
    if func_name is None:
        print("[Error] Failed to get function name")
        return False

    if objc.is_objc_method(func_name):
        if objc.is_objc_static_method(func_name):
            print("[Error] Static Obj-C method has no self")
            return False

        is_objc = True
        class_name = func_name.split(" ")[0][2:]
    else:
        # Try C++
        is_objc = False
        class_name = cpp.demangle_class_only(memory.name_from_ea(func.start_ea))
        if class_name is None:
            print("[Error] Failed to get class name in C++ mode")
            return False

    func_details = tif.get_func_details(func)
    if func_details is None:
        print("[Error] Failed to get function type info")
        return False

    if func_details.size() < 1:
        print("[Error] Function does not have enough arguments")
        return False

    # Change first argument name and type
    class_tinfo = tif.from_struct_name(class_name)
    if class_tinfo is None:
        print(f"[Error] Failed to get class type info for {class_name}")
        return False

    func_details[0].name = "self" if is_objc else "this"
    func_details[0].type = tif.pointer_of(class_tinfo)

    # Apply the changes
    new_tinfo = tif.from_func_details(func_details)
    if not tif.apply_tinfo_to_func(new_tinfo, func):
        print("[Error] Failed to apply new type info on function")
        return False

    print("Successfully updated first argument")
    return True



import contextlib
import re
from collections.abc import Callable, Hashable, Iterable, Iterator
from functools import wraps
from typing import Generic, TypeVar, overload


class CounterMixin:
    cnt: int = 0

    def count(self, amount: int = 1):
        self.cnt += amount


def match(arr: list[str | re.Pattern], item: str) -> bool:
    """Match a string against a list of strings or regex patterns."""
    for pat in arr:
        if isinstance(pat, str):
            if item == pat:
                return True
        else:
            if pat.match(item):
                return True
    return False


T = TypeVar("T")


def match_dict(patterns: dict[str | re.Pattern, T], item: str) -> T | None:
    """match a string against a dictionary of strings or regex patterns, Returns the value if matched."""
    for pat, val in patterns.items():
        if isinstance(pat, str):
            if item == pat:
                return val
        else:
            if pat.match(item):
                return val
    return None


K = TypeVar("K")
V = TypeVar("V")


class CustomDict(Generic[K, V]):
    def __init__(self, hasher: Callable[[K], Hashable]):
        self._hasher = hasher
        self._storage: dict[Hashable, tuple[K, V]] = {}

    def __setitem__(self, key: K, value: V):
        self._storage[self._hasher(key)] = (key, value)

    def __getitem__(self, key: K) -> V:
        return self._storage[self._hasher(key)][1]

    def __delitem__(self, key: K):
        del self._storage[self._hasher(key)]

    def __contains__(self, key: K) -> bool:
        return self._hasher(key) in self._storage

    def __str__(self) -> str:
        return f"{{ {', '.join(f'{key}: {value}' for key, value in self._storage.values())} }}"

    @overload
    def get(self, key: K) -> V | None: ...

    @overload
    def get(self, key: K, default: V) -> V: ...

    @overload
    def get(self, key: K, default: V | None) -> V | None: ...

    def get(self, key: K, default: V | None = None):
        if key in self:
            return self[key]
        return default

    def setdefault(self, key: K, default: V) -> V:
        cur = self.get(key)
        if cur is not None:
            return cur

        self[key] = default
        return default

    def __len__(self) -> int:
        return len(self._storage)

    def keys(self) -> Iterator[K]:
        return (k for k, _ in self._storage.values())

    def values(self) -> Iterator[V]:
        return (v for _, v in self._storage.values())

    def items(self) -> Iterator[tuple[K, V]]:
        return iter(self._storage.values())

    def __iter__(self) -> Iterator[K]:
        return self.keys()

    def __bool__(self) -> bool:
        return bool(self._storage)


class CustomSet(Generic[V]):
    def __init__(self, hasher: Callable[[V], Hashable]):
        self._hasher = hasher
        self._storage: dict[Hashable, V] = {}

    def add(self, value: V):
        self._storage[self._hasher(value)] = value

    def add_all(self, items: Iterator[V] | Iterable[V]):
        for item in items:
            self.add(item)

    def remove(self, value: V):
        del self._storage[self._hasher(value)]

    def discard(self, value: V):
        with contextlib.suppress(KeyError):
            self.remove(value)

    def __contains__(self, value: V) -> bool:
        return self._hasher(value) in self._storage

    def __len__(self) -> int:
        return len(self._storage)

    def __iter__(self) -> Iterator[V]:
        return iter(self._storage.values())

    def __bool__(self) -> bool:
        return bool(self._storage)

    def update(self, other: "CustomSet[V]") -> None:
        self._storage.update(other._storage)

    def __or__(self, other: "CustomSet[V]") -> "CustomSet[V]":
        new_set = CustomSet(self._hasher)
        new_set |= self
        new_set |= other
        return new_set

    def __ior__(self, other: "CustomSet[V]") -> "CustomSet[V]":
        self.update(other)
        return self

    def intersection_update(self, other: "CustomSet[V]") -> None:
        if not other:
            self._storage.clear()
            return

        self._storage = {k: self._storage[k] for k in (self._storage.keys() & other._storage.keys())}

    def __iand__(self, other: "CustomSet[V]") -> "CustomSet[V]":
        self.intersection_update(other)
        return self


def cache_fast(func: Callable[[], T]) -> Callable[[], T]:
    """Decorator to cache the result of a function for faster access."""
    cached_value: T | None = None

    @wraps(func)
    def wrapper():
        nonlocal cached_value
        if cached_value is None:
            cached_value = func()
        return cached_value

    return wrapper
